# Copyright (C) 2016 Kristoffer Gronlund <kgronlund@suse.com>
# See COPYING for license information.
#
# Bootstrap:
#
# Supersedes and replaces both the init/add/remove cluster scripts,
# and the ha-cluster-bootstrap scripts.
#
# Implemented as a straight-forward set of python functions for
# simplicity and flexibility.
#
# TODO: Make csync2 usage optional
# TODO: Configuration file for bootstrap?

import os
import sys
import random
import re
import time
import readline
from string import Template
from lxml import etree
from . import config
from . import utils
from . import xmlutil
from .cibconfig import mkset_obj, cib_factory
from . import corosync
from . import tmpfiles
from . import clidisplay
from . import term


LOG_FILE = "/var/log/ha-cluster-bootstrap.log"
CSYNC2_KEY = "/etc/csync2/key_hagroup"
CSYNC2_CFG = "/etc/csync2/csync2.cfg"
COROSYNC_AUTH = "/etc/corosync/authkey"
SYSCONFIG_SBD = "/etc/sysconfig/sbd"
SYSCONFIG_FW = "/etc/sysconfig/SuSEfirewall2"
SYSCONFIG_FW_CLUSTER = "/etc/sysconfig/SuSEfirewall2.d/services/cluster"
PCMK_REMOTE_AUTH = "/etc/pacemaker/authkey"


INIT_STAGES = ("ssh", "ssh_remote", "csync2", "csync2_remote", "corosync", "storage", "sbd", "cluster", "vgfs", "admin")


class Context(object):
    """
    Context object used to avoid having to pass these variables
    to every bootstrap method.
    """
    def __init__(self, quiet, yes_to_all, nic=None, ip_address=None, ip_network=None):
        self.quiet = quiet
        self.yes_to_all = yes_to_all
        self.nic = nic
        self.ip_address = ip_address
        self.ip_network = ip_network
        self.cluster_name = None
        self.cluster_node = None
        self.ocfs2_device = None
        self.shared_device = None
        self.sbd_device = None
        self.diskless_sbd = False  # if True, enable SBD for diskless operation
        self.unicast = None
        self.ipv6 = None
        self.admin_ip = None
        self.watchdog = None
        self.host_status = None
        self.connect_name = None
        self.second_hb = None
        self.ui_context = None
        self.qdevice = None


_context = None


def die(*args):
    """
    Broken out as special case for log() failure.  Ordinarily you
    should just use error() to terminate.
    """
    raise ValueError(" ".join([str(arg) for arg in args]))


def error(*args):
    """
    Log an error message and raise ValueError to bail out of
    bootstrap process.
    """
    log("ERROR: {}".format(" ".join([str(arg) for arg in args])))
    die(*args)


def warn(*args):
    """
    Log and display a warning message.
    """
    log("WARNING: {}".format(" ".join(str(arg) for arg in args)))
    if not _context.quiet:
        print(term.render(clidisplay.warn("! {}".format(" ".join(str(arg) for arg in args)))))


@utils.memoize
def log_file_fallback():
    """
    If the standard log location isn't writable,
    just log to the nearest temp dir.
    """
    return os.path.join(utils.get_tempdir(), "ha-cluster-bootstrap.log")


def log(*args):
    global LOG_FILE
    try:
        with open(LOG_FILE, "ab") as logfile:
            text = " ".join([utils.to_ascii(arg) for arg in args]) + "\n"
            logfile.write(text.encode('ascii', 'backslashreplace'))
    except IOError:
        if LOG_FILE != log_file_fallback():
            LOG_FILE = log_file_fallback()
            log(*args)
        else:
            die("Can't append to {} - aborting".format(LOG_FILE))


def drop_last_history():
    hlen = readline.get_current_history_length()
    if hlen > 0:
        readline.remove_history_item(hlen - 1)


def prompt_for_string(msg, match=None, default='', valid_func=None, prev_value=None):
    if _context.yes_to_all:
        return default
    while True:
        disable_completion()
        val = utils.multi_input('  %s [%s]' % (msg, default))
        enable_completion()
        if not val:
            val = default
        else:
            drop_last_history()
        if match is None:
            return val
        if val and re.match(match, val) is not None:
            if not valid_func or valid_func(val, prev_value):
                return val
        print(term.render(clidisplay.error("    Invalid value entered")))


def confirm(msg):
    if _context.yes_to_all:
        return True
    disable_completion()
    rc = utils.ask(msg)
    enable_completion()
    drop_last_history()
    return rc


def disable_completion():
    if _context.ui_context:
        _context.ui_context.disable_completion()


def enable_completion():
    if _context.ui_context:
        _context.ui_context.setup_readline()


def invoke(*args):
    """
    Log command execution to log file.
    Log output from command to log file.
    Return returncode == 0
    """
    log("+ " + " ".join(args))
    rc, stdout, stderr = utils.get_stdout_stderr(" ".join(args))
    if stdout:
        log(stdout)
    if stderr:
        log(stderr)
    return rc == 0


def invokerc(*args):
    "Like invoke, but returns the returncode, not True/False"
    log("+ " + " ".join(args))
    rc, stdout, stderr = utils.get_stdout_stderr(" ".join(args))
    if stdout:
        log(stdout)
    if stderr:
        log(stderr)
    return rc


def crm_configure_load(action, configuration):
    log(": loading crm config (%s), content is:" % (action))
    log(configuration)
    if not cib_factory.initialize():
        error("Failed to load cluster configuration")
    set_obj = mkset_obj()
    if action == 'replace':
        cib_factory.erase()
    if not set_obj.save(configuration, remove=False, method=action):
        error("Failed to load cluster configuration")
    if not cib_factory.commit():
        error("Failed to commit cluster configuration")


def wait_for_resource(message, resource, needle="running on"):
    status_long(message)
    while True:
        _rc, out, err = utils.get_stdout_stderr("crm_resource --locate --resource " + resource)
        if needle in out:
            break
        if needle in err:
            break
        status_progress()
        sleep(1)
    status_done()


def wait_for_stop(message, resource):
    return wait_for_resource(message, resource, needle="NOT running")


def wait_for_cluster():
    status_long("Waiting for cluster")
    while True:
        _rc, out, _err = utils.get_stdout_stderr("crm_mon -1")
        if "online" in out.lower():
            break
        status_progress()
        sleep(5)
    status_done()


def start_service(service):
    """
    Start and enable systemd service
    """
    invoke("systemctl enable " + service)
    rc, _out, _err = utils.get_stdout_stderr('systemctl -q is-active ' + service)
    if rc != 0:
        if not invoke("systemctl start " + service):
            error("Failed to start " + service)


def stop_service(service):
    """
    Stops the systemd service
    """
    return invoke("systemctl stop " + service)


def service_is_enabled(service):
    """
    Check if service is enabled
    """
    rc, _out, _err = utils.get_stdout_stderr('systemctl is-enabled %s' % (service))
    return rc == 0


def service_is_active(service):
    """
    Check if service is active
    """
    rc, _out, _err = utils.get_stdout_stderr('systemctl -q is-active %s' % (service))
    return rc == 0


def package_is_installed(pkg):
    """
    Check if package is installed
    """
    return invoke("rpm -q --quiet {}".format(pkg))


def sleep(t):
    """
    Sleep for t seconds.
    """
    t = float(t)
    time.sleep(t)


def status(msg):
    log("# " + msg)
    if not _context.quiet:
        print("  {}".format(msg))


def status_long(msg):
    log("# {}...".format(msg))
    if not _context.quiet:
        sys.stdout.write("  {}...".format(msg))
        sys.stdout.flush()


def status_progress():
    if not _context.quiet:
        sys.stdout.write(".")
        sys.stdout.flush()


def status_done():
    log("# done")
    if not _context.quiet:
        print("done")


def partprobe():
    # This function uses fdisk to create a list of valid devices for probing
    # with partprobe.  This prevents partprobe from failing on read-only mounted
    # devices such as /dev/sr0 (etc) that might cause it to return an error when
    # it exits.  This allows partprobe to run without forcing _die to bail out.
    # -Brandon Heaton
    #  ATT Training Engineer
    #  Data Center Engineer
    #  bheaton@suse.com
    _rc, out, _err = utils.get_stdout_stderr("sfdisk -l")
    disks = re.findall(r'^Disk\s*(/.+):', out, re.M)
    invoke("partprobe", *disks)


def probe_partitions():
    status_long("Probing for new partitions")
    partprobe()
    sleep(5)
    status_done()


def configured_sbd_device():
    """
    Gets currently configured SBD device, i.e. what's in /etc/sysconfig/sbd
    """
    conf = utils.parse_sysconfig(SYSCONFIG_SBD)
    return conf.get("SBD_DEVICE")


def check_tty():
    """
    Check for pseudo-tty: Cannot display read prompts without a TTY (bnc#892702)
    """
    if _context.yes_to_all:
        return
    if not sys.stdin.isatty():
        error("No pseudo-tty detected! Use -t option to ssh if calling remotely.")


def grep_output(cmd, txt):
    _rc, outp, _err = utils.get_stdout_stderr(cmd)
    return txt in outp


def grep_file(fn, txt):
    return os.path.exists(fn) and txt in open(fn).read()


def service_is_available(svcname):
    return grep_output("systemctl list-unit-files {}".format(svcname), svcname)


def my_hostname_resolves():
    import socket
    hostname = utils.this_node()
    try:
        socket.gethostbyname(hostname)
        return True
    except socket.error:
        return False


def check_prereqs(stage):
    warned = False

    if not my_hostname_resolves():
        warn("Hostname '{}' is unresolvable. {}".format(
            utils.this_node(),
            "Please add an entry to /etc/hosts or configure DNS."))
        warned = True

    timekeepers = ('chronyd.service', 'ntp.service', 'ntpd.service')
    timekeeper = None
    for tk in timekeepers:
        if service_is_available(tk):
            timekeeper = tk
            break

    if timekeeper is None:
        warn("No NTP service found.")
        warned = True
    elif not service_is_enabled(timekeeper):
        warn("{} is not configured to start at system boot.".format(timekeeper))
        warned = True

    if stage in ("", "join", "sbd"):
        if not check_watchdog():
            warn("No watchdog device found. If SBD is used, the cluster will be unable to start without a watchdog.")
            warned = True

    if warned:
        if not confirm("Do you want to continue anyway?"):
            return False

    firewall_open_basic_ports()
    return True


def init_watchdog():
    """
    If a watchdog device was provided on the command line, configure it
    """
    if _context.watchdog:
        _, outp = utils.get_stdout("lsmod")
        if not any(re.match(r"^{}\b", d) is not None for d in outp.splitlines()):
            invoke("modprobe {}".format(_context.watchdog))
            with open("/etc/modules-load.d/watchdog.conf", "w") as f:
                f.write("{}\n".format(_context.watchdog))
            invoke("systemctl restart systemd-modules-load")


def log_start():
    """
    Convenient side-effect: this will die immediately if the log file
    is not writable (e.g. if not running as root)
    """
    # Reload rsyslog to make sure it logs with the correct hostname
    if service_is_active("rsyslog.service"):
        invoke("systemctl reload rsyslog.service")
    datestr = utils.get_stdout("date --rfc-3339=seconds")[1]
    log('================================================================')
    log("%s %s" % (datestr, " ".join(sys.argv)))
    log('----------------------------------------------------------------')


def init_network():
    """
    Auto-detection of IP address and netmask only works if $NET_IF is
    up. If $IP_ADDRESS is overridden, network detect shouldn't work,
    because "ip route" won't be able to help us.
    """
    if not _context.ipv6:
        nic, ipaddr, ipnetwork, _prefix = utils.network_defaults(_context.nic)
    else:
        all_ = utils.network_v6_all()
        if not all_:
            error("Unable to configure network: No usable IPv6 addresses found.")
        nic = sorted(all_.keys())[0]
        ipaddr = all_[nic][0].split('/')[0]
        ipnetwork = utils.get_ipv6_network(all_[nic][0])

    if _context.nic is None:
        _context.nic = nic
    if _context.ip_address is None:
        _context.ip_address = ipaddr
    if _context.ip_network is None:
        _context.ip_network = ipnetwork
    if _context.ip_address is None:
        warn("Could not detect IP address for {}".format(nic))
    if _context.ip_network is None:
        warn("Could not detect IP network address for {}".format(nic))


def configure_firewall(tcp=None, udp=None):
    if tcp is None:
        tcp = []
    if udp is None:
        udp = []

    def init_firewall_suse(tcp, udp):
        if os.path.exists(SYSCONFIG_FW_CLUSTER):
            cluster = utils.parse_sysconfig(SYSCONFIG_FW_CLUSTER)
            tcpcurr = set(cluster.get("TCP", "").split())
            tcpcurr.update(tcp)
            tcp = list(tcpcurr)
            udpcurr = set(cluster.get("UDP", "").split())
            udpcurr.update(udp)
            udp = list(udpcurr)

        utils.sysconfig_set(SYSCONFIG_FW_CLUSTER, TCP=" ".join(tcp), UDP=" ".join(udp))

        ext = ""
        if os.path.exists(SYSCONFIG_FW):
            fw = utils.parse_sysconfig(SYSCONFIG_FW)
            ext = fw.get("FW_CONFIGURATIONS_EXT", "")
            if "cluster" not in ext.split():
                ext = ext + " cluster"
        utils.sysconfig_set(SYSCONFIG_FW, FW_CONFIGURATIONS_EXT=ext)

        # No need to do anything else if the firewall is inactive
        if not service_is_active("SuSEfirewall2"):
            return

        # Firewall is active, either restart or complain if we couldn't tweak it
        status("Restarting firewall (tcp={}, udp={})".format(" ".join(tcp), " ".join(udp)))
        if not invoke("rcSuSEfirewall2 restart"):
            error("Failed to restart firewall (SuSEfirewall2)")

    def init_firewall_firewalld(tcp, udp):
        has_firewalld = service_is_active("firewalld")
        cmdbase = 'firewall-cmd --zone=public --permanent ' if has_firewalld else 'firewall-offline-cmd --zone=public '

        def cmd(args):
            if not invoke(cmdbase + args):
                error("Failed to configure firewall.")

        for p in tcp:
            cmd("--add-port={}/tcp".format(p))

        for p in udp:
            cmd("--add-port={}/udp".format(p))

        if has_firewalld:
            if not invoke("firewall-cmd --reload"):
                error("Failed to reload firewall configuration.")

    def init_firewall_ufw(tcp, udp):
        """
        try configuring firewall with ufw
        """
        for p in tcp:
            if not invoke("ufw allow {}/tcp".format(p)):
                error("Failed to configure firewall (ufw)")
        for p in udp:
            if not invoke("ufw allow {}/udp".format(p)):
                error("Failed to configure firewall (ufw)")

    if package_is_installed("firewalld"):
        init_firewall_firewalld(tcp, udp)
    elif package_is_installed("SuSEfirewall2"):
        init_firewall_suse(tcp, udp)
    elif package_is_installed("ufw"):
        init_firewall_ufw(tcp, udp)
    else:
        warn("Failed to detect firewall: Could not open ports tcp={}, udp={}".format("|".join(tcp), "|".join(udp)))


def firewall_open_basic_ports():
    """
    Open ports for csync2, mgmtd, hawk & dlm respectively
    """
    configure_firewall(tcp=["30865", "5560", "7630", "21064"])


def firewall_open_corosync_ports():
    """
    Have to do this separately, as we need general firewall config early
    so csync2 works, but need corosync config *after* corosync.conf has
    been created/updated.

    Please note corosync uses two UDP ports mcastport (for mcast
    receives) and mcastport - 1 (for mcast sends).

    Also open QNetd/QDevice port if configured.
    """
    # all mcastports defined in corosync config
    udp = corosync.get_values("totem.interface.mcastport")
    udp.extend([str(int(p) - 1) for p in udp])

    tcp = corosync.get_values("totem.quorum.device.net.port")

    configure_firewall(tcp=tcp, udp=udp)


def init_cluster_local():
    # Caller should check this, but I'm paranoid...
    if service_is_active("corosync.service"):
        error("corosync service is running!")

    firewall_open_corosync_ports()

    # reset password, but only if it's not already set
    _rc, outp = utils.get_stdout("passwd -S hacluster")
    ps = outp.strip().split()[1]
    pass_msg = ""
    if ps not in ("P", "PS"):
        log(': Resetting password of hacluster user')
        rc, outp, errp = utils.get_stdout_stderr("passwd hacluster", input_s=b"linux\nlinux\n")
        if rc != 0:
            warn("Failed to reset password of hacluster user: %s" % (outp + errp))
        else:
            pass_msg = ", password 'linux'"

    # evil, but necessary
    invoke("rm -f /var/lib/heartbeat/crm/* /var/lib/pacemaker/cib/*")

    # only try to start hawk if hawk is installed
    if service_is_available("hawk.service"):
        start_service("hawk.service")
        status("Hawk cluster interface is now running. To see cluster status, open:")
        status("  https://%s:7630/" % (_context.ip_address))
        status("Log in with username 'hacluster'%s" % (pass_msg))
    else:
        warn("Hawk not installed - not configuring web management interface.")

    if pass_msg:
        warn("You should change the hacluster password to something more secure!")

    # for cluster join, diskless_sbd flag is set in join_cluster() if
    # sbd is running on seed host
    if configured_sbd_device() or _context.diskless_sbd:
        invoke("systemctl enable sbd.service")
    else:
        invoke("systemctl disable sbd.service")

    start_service("pacemaker.service")
    wait_for_cluster()


def install_tmp(tmpfile, to):
    with open(tmpfile, "r") as src:
        with utils.open_atomic(to, "w") as dst:
            for line in src:
                dst.write(line)


def append(fromfile, tofile):
    log("+ cat %s >> %s" % (fromfile, tofile))
    with open(tofile, "a") as tf:
        with open(fromfile, "r") as ff:
            tf.write(ff.read())


def rmfile(path, ignore_errors=False):
    """
    Try to remove the given file, and
    report an error on failure
    """
    try:
        os.remove(path)
    except os.error as err:
        if not ignore_errors:
            error("Failed to remove {}: {}".format(path, err))


def mkdirs_owned(dirs, mode=0o777, uid=-1, gid=-1):
    """
    Create directory path, setting the mode and
    ownership of the leaf directory to mode/uid/gid.
    """
    if not os.path.exists(dirs):
        try:
            os.makedirs(dirs, mode)
        except OSError as err:
            error("Failed to create {}: {}".format(dirs, err))
        if uid != -1 or gid != -1:
            utils.chown(dirs, uid, gid)


def init_api_server():
    invoke("clusterAPIServer &> /dev/null &")


def init_ssh():
    """
    Configure passwordless SSH.
    """
    start_service("sshd.service")
    invoke("mkdir -m 700 -p /root/.ssh")
    if os.path.exists("/root/.ssh/id_rsa"):
        if not confirm("/root/.ssh/id_rsa already exists - overwrite?"):
            return
        rmfile("/root/.ssh/id_rsa")
    status("Generating SSH key")
    invoke("ssh-keygen -q -f /root/.ssh/id_rsa -C 'Cluster Internal' -N ''")
    append("/root/.ssh/id_rsa.pub", "/root/.ssh/authorized_keys")


def init_ssh_remote():
    """
    Called by ha-cluster-join
    """
    authkeys = open("/root/.ssh/authorized_keys", "r+")
    authkeys_data = authkeys.read()
    for key in ("id_rsa", "id_dsa", "id_ecdsa", "id_ed25519"):
        fn = os.path.join("/root/.ssh", key)
        if not os.path.exists(fn):
            continue
        keydata = open(fn + ".pub").read()
        if keydata not in authkeys_data:
            append(fn + ".pub", "/root/.ssh/authorized_keys")


def init_csync2():
    status("Configuring csync2")
    if os.path.exists(CSYNC2_KEY):
        if not confirm("csync2 is already configured - overwrite?"):
            return

    invoke("rm", "-f", CSYNC2_KEY)
    status_long("Generating csync2 shared key (this may take a while)")
    if not invoke("csync2", "-k", CSYNC2_KEY):
        error("Can't create csync2 key {}".format(CSYNC2_KEY))
    status_done()

    utils.str2file("""group ha_group
{
key /etc/csync2/key_hagroup;
host %s;
include /etc/booth;
include /etc/corosync/corosync.conf;
include /etc/corosync/authkey;
include /etc/csync2/csync2.cfg;
include /etc/csync2/key_hagroup;
include /etc/ctdb/nodes;
include /etc/drbd.conf;
include /etc/drbd.d;
include /etc/ha.d/ldirectord.cf;
include /etc/lvm/lvm.conf;
include /etc/multipath.conf;
include /etc/samba/smb.conf;
include /etc/sysconfig/pacemaker;
include /etc/sysconfig/sbd;
include /etc/pacemaker/authkey;
}
    """ % (utils.this_node()), CSYNC2_CFG)

    start_service("csync2.socket")
    status_long("csync2 checking files")
    invoke("csync2", "-cr", "/")
    status_done()


def csync2_update(path):
    invoke("csync2 -rm %s" % (path))
    invoke("csync2 -rf %s" % (path))
    invoke("csync2 -rxv %s" % (path))


def init_csync2_remote():
    """
    It would be nice if we could just have csync2.cfg include a directory,
    which in turn included one file per node which would be referenced via
    something like "group ha_group { ... config: /etc/csync2/hosts/*; }"
    That way, adding a new node would just mean adding a single new file
    to that directory.  Unfortunately, the 'config' statement only allows
    inclusion of specific individual files, not multiple files via wildcard.
    So we have this function which is called by ha-cluster-join to add the new
    remote node to csync2 config on some existing node.  It is intentionally
    not documented in ha-cluster-init's user-visible usage information.
    """
    newhost = _context.cluster_node
    if not newhost:
        error("Hostname not specified")

    curr_cfg = open(CSYNC2_CFG).read()

    was_quiet = _context.quiet
    try:
        _context.quiet = True
        # if host doesn't already exist in csync2 config, add it
        if not re.search(r"^\s*host.*\s+%s\s*;" % (newhost), curr_cfg, flags=re.M):
            curr_cfg = re.sub(r"\bhost.*\s+\S+\s*;", r"\g<0>\n\thost %s;" % (utils.doublequote(newhost)), curr_cfg, count=1)
            utils.str2file(curr_cfg, CSYNC2_CFG)
            csync2_update("/")
        else:
            log(": Not updating %s - remote host %s already exists" % (CSYNC2_CFG, newhost))
    finally:
        _context.quiet = was_quiet


def init_corosync_auth():
    """
    Generate the corosync authkey
    """
    if os.path.exists(COROSYNC_AUTH):
        if not confirm("%s already exists - overwrite?" % (COROSYNC_AUTH)):
            return
        rmfile(COROSYNC_AUTH)
    invoke("corosync-keygen -l")


def init_remote_auth():
    """
    Generate the pacemaker-remote authkey
    """
    if os.path.exists(PCMK_REMOTE_AUTH):
        if not confirm("%s already exists - overwrite?" % (PCMK_REMOTE_AUTH)):
            return
        rmfile(PCMK_REMOTE_AUTH)

    pcmk_remote_dir = os.path.dirname(PCMK_REMOTE_AUTH)
    mkdirs_owned(pcmk_remote_dir, mode=0o750, gid="haclient")
    if not invoke("dd if=/dev/urandom of={} bs=4096 count=1".format(PCMK_REMOTE_AUTH)):
        warn("Failed to create pacemaker authkey: {}".format(PCMK_REMOTE_AUTH))
    utils.chown(PCMK_REMOTE_AUTH, "hacluster", "haclient")
    os.chmod(PCMK_REMOTE_AUTH, 0o640)


def valid_network(addr, prev_value=None):
    """
    bindnetaddr(IPv4) must be one of the local networks
    """
    if prev_value and addr == prev_value[0]:
        warn("  Network address '{}' is already in use!".format(addr))
        return False
    all_ = utils.network_all()
    if all_ and addr in all_:
        return True
    warn("  Address '{}' invalid, expected one of {}".format(addr, all_))
    return False


def valid_v6_network(addr, prev_value=None):
    """
    bindnetaddr(IPv6) must be one of the local networks
    """
    if prev_value and addr == prev_value[0]:
        warn("  Network address '{}' is already in use!".format(addr))
        return False
    network_list = []
    all_ = utils.network_v6_all()
    # the format of return example:
    # {'eth2': ['2002:db8::3/64'], 'eth1': ['2001:db8::3/64']}
    if not all_:
        return False
    for item in list(all_.values()):
        network_list.extend(item)
    network_list = [utils.get_ipv6_network(x) for x in network_list]
    if addr in network_list:
        return True
    warn("  Address '{}' invalid, expected one of {}".format(addr, network_list))
    return False


def valid_ucastIP(addr, prev_value=None):
    if prev_value and addr == prev_value[0]:
        print(term.render(clidisplay.error("    IP {} is already in use".format(addr))))
        return False
    ip_local = utils.ip_in_local(_context.ipv6)
    if addr not in ip_local:
        print(term.render(clidisplay.error("    IP must be a local address (one of {})".format(ip_local))))
        return False
    return True


def valid_adminIP(addr, prev_value=None):
    """
    valid adminIP for IPv4/IPv6
    """
    try:
        ip = utils.IP(addr)
    except ValueError as err:
        print(term.render(clidisplay.error("    {}".format(err))))
        return False
    else:
        all_ = []
        if ip.version() == 4:
            ping = "ping"
            all_ = utils.network_all(with_mask=True)
        else:
            # for IPv6
            ping = "ping6"
            network_list = utils.network_v6_all()
            for item in list(network_list.values()):
                all_.extend(item)
        if invoke("{} -c 1 {}".format(ping, addr)):
            warn("  Address already in use: {}".format(addr))
            return False
        for net in all_:
            if addr in utils.Network(net):
                return True
        warn("  Address '{}' invalid, expected one of {}".format(addr, all_))
        return False


def valid_ipv4_addr(addr, prev_value=None):
    if prev_value and addr == prev_value[0]:
        warn("  Address already in use: {}".format(addr))
        return False
    return utils.valid_ip_addr(addr)


def valid_ipv6_addr(addr, prev_value=None):
    if prev_value and addr == prev_value[0]:
        warn("  Address already in use: {}".format(addr))
        return False
    return utils.valid_ip_addr(addr, 6)


def valid_port(port, prev_value=None):
    if prev_value:
        if abs(int(port) - int(prev_value[0])) <= 1:
            warn("  Port {} is already in use by corosync. Leave a gap between multiple rings.".format(port))
            return False
    if int(port) >= 1024 and int(port) <= 65535:
        return True
    return False


def init_corosync_unicast():
    def pick_default_value(vlist, ilist):
        # give a different value for second config items
        if len(ilist) == 1:
            vlist.remove(ilist[0])
        return vlist[0]

    if _context.yes_to_all:
        status("Configuring corosync (unicast)")
    else:
        status("""
Configure Corosync (unicast):
  This will configure the cluster messaging layer.  You will need
  to specify a network address over which to communicate (default
  is %s's network, but you can use the network address of any
  active interface).
""" % (_context.nic))

    if os.path.exists(corosync.conf()):
        if not confirm("%s already exists - overwrite?" % (corosync.conf())):
            return

    ringXaddr_res = []
    mcastport_res = []
    default_ports = ["5405", "5407"]
    two_rings = False
    default_networks = []

    if _context.ipv6:
        network_list = []
        all_ = utils.network_v6_all()
        for item in all_.values():
            network_list.extend(item)
        default_networks = map(utils.get_ipv6_network, network_list)
    else:
        default_networks = utils.network_all()
    if not default_networks:
        error("No network configured at {}!".format(utils.this_node()))

    for i in 0, 1:
        ringXaddr = prompt_for_string('Address for ring{}'.format(i),
                                      r'([0-9]+\.){3}[0-9]+|[0-9a-fA-F]{1,4}:',
                                      _context.ip_address if i == 0 and _context.ip_address else "",
                                      valid_ucastIP,
                                      ringXaddr_res)
        if not ringXaddr:
            error("No value for ring{}".format(i))
        ringXaddr_res.append(ringXaddr)

        mcastport = prompt_for_string('Port for ring{}'.format(i),
                                      '[0-9]+',
                                      pick_default_value(default_ports, mcastport_res),
                                      valid_port,
                                      mcastport_res)
        if not mcastport:
            error("Expected a multicast port for ring{}".format(i))
        mcastport_res.append(mcastport)

        if i == 1 or \
           len(default_networks) == 1 or \
           not _context.second_hb or \
           not confirm("\nAdd another heartbeat line?"):
            break
        two_rings = True

    corosync.create_configuration(
        clustername=_context.cluster_name,
        ringXaddr=ringXaddr_res,
        mcastport=mcastport_res,
        transport="udpu",
        ipv6=_context.ipv6,
        two_rings=two_rings,
        qdevice=_context.qdevice)
    csync2_update(corosync.conf())


def init_corosync_multicast():
    def gen_mcastaddr():
        if _context.ipv6:
            return "ff3e::%s:%d" % (
                ''.join([random.choice('0123456789abcdef') for _ in range(4)]),
                random.randint(0, 9))
        return "239.%d.%d.%d" % (
            random.randint(0, 255),
            random.randint(0, 255),
            random.randint(1, 255))

    def pick_default_value(vlist, ilist):
        # give a different value for second config items
        if len(ilist) == 1:
            vlist.remove(ilist[0])
        return vlist[0]

    if _context.yes_to_all:
        status("Configuring corosync")
    else:
        status("""
Configure Corosync:
  This will configure the cluster messaging layer.  You will need
  to specify a network address over which to communicate (default
  is %s's network, but you can use the network address of any
  active interface).
""" % (_context.nic))

    if os.path.exists(corosync.conf()):
        if not confirm("%s already exists - overwrite?" % (corosync.conf())):
            return

    bindnetaddr_res = []
    mcastaddr_res = []
    mcastport_res = []
    default_ports = ["5405", "5407"]
    two_rings = False
    default_networks = []

    if _context.ipv6:
        network_list = []
        all_ = utils.network_v6_all()
        for item in list(all_.values()):
            network_list.extend(item)
        default_networks = [utils.get_ipv6_network(x) for x in network_list]
    else:
        default_networks = utils.network_all()
    if not default_networks:
        error("No network configured at {}!".format(utils.this_node()))

    for i in 0, 1:
        if _context.ipv6:
            bindnetaddr = prompt_for_string('Network address to bind to',
                                            r'.*(::|0)$',
                                            pick_default_value(default_networks, bindnetaddr_res),
                                            valid_v6_network,
                                            bindnetaddr_res)
            if not bindnetaddr:
                error("No value for bindnetaddr")
            bindnetaddr_res.append(bindnetaddr)

            mcastaddr = prompt_for_string('Multicast address',
                                          r'^[Ff][Ff].*',
                                          gen_mcastaddr(),
                                          valid_ipv6_addr,
                                          mcastaddr_res)
            if not mcastaddr:
                error("No value for mcastaddr")
            mcastaddr_res.append(mcastaddr)

        else:
            bindnetaddr = prompt_for_string('Network address to bind to (e.g.: 192.168.1.0)',
                                            r'([0-9]+\.){3}0$',
                                            pick_default_value(default_networks, bindnetaddr_res),
                                            valid_network,
                                            bindnetaddr_res)
            if not bindnetaddr:
                error("No value for bindnetaddr")
            bindnetaddr_res.append(bindnetaddr)

            mcastaddr = prompt_for_string('Multicast address (e.g.: 239.x.x.x)',
                                          utils.mcast_regrex,
                                          gen_mcastaddr(),
                                          valid_ipv4_addr,
                                          mcastaddr_res)
            if not mcastaddr:
                error("No value for mcastaddr")
            mcastaddr_res.append(mcastaddr)

        mcastport = prompt_for_string('Multicast port',
                                      '[0-9]+',
                                      pick_default_value(default_ports, mcastport_res),
                                      valid_port,
                                      mcastport_res)
        if not mcastport:
            error("No value for mcastport")
        mcastport_res.append(mcastport)

        if i == 1 or \
           len(default_networks) == 1 or \
           not _context.second_hb or \
           not confirm("\nConfigure a second multicast ring?"):
            break
        two_rings = True

    nodeid = None
    if _context.ipv6:
        nodeid = utils.gen_nodeid_from_ipv6(_context.ip_address)

    corosync.create_configuration(
        clustername=_context.cluster_name,
        bindnetaddr=bindnetaddr_res,
        mcastaddr=mcastaddr_res,
        mcastport=mcastport_res,
        ipv6=_context.ipv6,
        nodeid=nodeid,
        two_rings=two_rings,
        qdevice=_context.qdevice)
    #csync2_update(corosync.conf())


def init_corosync():
    """
    Configure corosync (unicast or multicast, encrypted?)
    """
    @utils.memoize
    def check_amazon():
        if not utils.is_program("dmidecode"):
            return False
        _rc, outp = utils.get_stdout("dmidecode -s system-version")
        return re.search(r"\<.*\.amazon\>", outp) is not None

    init_corosync_auth()
    if _context.unicast or check_amazon():
        init_corosync_unicast()
    else:
        init_corosync_multicast()


def is_block_device(dev):
    from stat import S_ISBLK
    try:
        rc = S_ISBLK(os.stat(dev).st_mode)
    except OSError as msg:
        warn(msg)
        return False
    return rc


def list_partitions(dev):
    rc, outp, errp = utils.get_stdout_stderr("parted -s %s print" % (dev))
    partitions = []
    for line in outp.splitlines():
        m = re.match(r"^\s*([0-9]+)\s*", line)
        if m:
            partitions.append(m.group(1))
    if rc != 0:
        # ignore "Error: /dev/vdb: unrecognised disk label"
        if errp.count('\n') > 1 or "unrecognised disk label" not in errp.strip():
            error("Failed to list partitions in {}: {}".format(dev, errp))
    return partitions


def list_devices(dev):
    "TODO: THIS IS *WRONG* FOR MULTIPATH! (but possibly nothing we can do about it)"
    _rc, outp = utils.get_stdout("fdisk -l %s" % (dev))
    partitions = []
    for line in outp.splitlines():
        m = re.match(r"^(\/dev\S+)", line)
        if m:
            partitions.append(m.group(1))
    return partitions


def init_storage():
    """
    Configure SBD and OCFS2 both on the same storage device.
    """
    dev = _context.shared_device
    partitions = []
    dev_looks_sane = False

    if _context.yes_to_all or not dev:
        status("Configuring shared storage")
    else:
        status("""
Configure Shared Storage:
  You will need to provide the path to a shared storage device,
  for example a SAN volume or iSCSI target.  The device path must
  be persistent and consistent across all nodes in the cluster,
  so /dev/disk/by-id/* devices are a good choice.  This device
  will be automatically paritioned into two pieces, 1MB for SBD
  fencing, and the remainder for an OCFS2 filesystem.
""")

    while not dev_looks_sane:
        dev = prompt_for_string('Path to storage device (e.g. /dev/disk/by-id/...)', r'\/.*', dev)
        if not dev:
            error("No value for shared storage device")

        if not is_block_device(dev):
            if _context.yes_to_all:
                error(dev + " is not a block device")
            else:
                print("    That doesn't look like a block device", file=sys.stderr)
        else:
            #
            # Got something that looks like a block device, there
            # are four possibilities now:
            #
            #  1) It's completely broken/inaccessible
            #  2) No recognizable partition table
            #  3) Empty partition table
            #  4) Non-empty parition table
            #
            partitions = list_partitions(dev)
            if partitions:
                status("WARNING: Partitions exist on %s!" % (dev))
                if confirm("Are you ABSOLUTELY SURE you want to overwrite?"):
                    dev_looks_sane = True
                else:
                    dev = ""
            else:
                # It's either broken, no partition table, or empty partition table
                status("%s appears to be empty" % (dev))
                if confirm("Device appears empty (no partition table). Do you want to use {}?".format(dev)):
                    dev_looks_sane = True
                else:
                    dev = ""

    if partitions:
        if not confirm("Really?"):
            return
        status_long("Erasing existing partitions...")
        for part in partitions:
            if not invoke("parted -s %s rm %s" % (dev, part)):
                error("Failed to remove partition %s from %s" % (part, dev))
        status_done()

    status_long("Creating partitions...")
    if not invoke("parted", "-s", dev, "mklabel", "msdos"):
        error("Failed to create partition table")

    # This is a bit rough, and probably won't result in great performance,
    # but it's fine for test/demo purposes to carve off 1MB for SBD.  Note
    # we have to specify the size of the first partition in this in bytes
    # rather than MB, or parted's rounding gives us a ~30Kb partition
    # (see rhbz#623268).
    if not invoke("parted -s %s mkpart primary 0 1048576B" % (dev)):
        error("Failed to create first partition on %s" % (dev))
    if not invoke("parted -s %s mkpart primary 1M 100%%" % (dev)):
        error("Failed to create second partition")

    status_done()

    # TODO: May not be strictly necessary, but...
    probe_partitions()

    # TODO: THIS IS *WRONG* FOR MULTIPATH! (but possibly nothing we can do about it)
    devices = list_devices(dev)

    _context.sbd_device = devices[0]
    if not _context.sbd_device:
        error("Unable to determine device path for SBD partition")

    _context.ocfs2_device = devices[1]
    if not _context.ocfs2_device:
        error("Unable to determine device path for OCFS2 partition")

    status("Created %s for SBD partition" % (_context.sbd_device))
    status("Created %s for OCFS2 partition" % (_context.ocfs2_device))


def detect_watchdog_device():
    """
    Find the watchdog device. Fall back to /dev/watchdog.
    """
    wdconf = "/etc/modules-load.d/watchdog.conf"
    watchdog_dev = "/dev/watchdog"
    if os.path.exists(wdconf):
        txt = open(wdconf, "r").read()
        for line in txt.splitlines():
            m = re.match(r'^\s*watchdog-device\s*=\s*(.*)$', line)
            if m:
                watchdog_dev = m.group(1)
    return watchdog_dev


def check_watchdog():
    """
    Verify watchdog device. Fall back to /dev/watchdog.
    """
    watchdog_dev = detect_watchdog_device()
    rc, _out, _err = utils.get_stdout_stderr("wdctl %s" % (watchdog_dev))
    return rc == 0


def sysconfig_comment_out(scfile, key):
    """
    Comments out the given key in the sysconfig file
    """
    matcher = re.compile(r'^\s*{}\s*='.format(key))
    outp, ncomments = "", 0
    for line in scfile.readlines():
        if matcher.match(line):
            outp += '#' + line
            ncomments += 1
        else:
            outp += line
    return outp, ncomments


def init_sbd_diskless():
    """
    Initialize SBD in diskless mode.
    """
    status_long("Initializing diskless SBD...")
    if os.path.isfile(SYSCONFIG_SBD):
        log("Overwriting {} with diskless configuration".format(SYSCONFIG_SBD))
        scfg, nmatches = sysconfig_comment_out(open(SYSCONFIG_SBD), "SBD_DEVICE")
        if nmatches > 0:
            utils.str2file(scfg, SYSCONFIG_SBD)
    else:
        log("Creating {} with diskless configuration".format(SYSCONFIG_SBD))
    utils.sysconfig_set(SYSCONFIG_SBD,
                        SBD_PACEMAKER="yes",
                        SBD_STARTMODE="always",
                        SBD_DELAY_START="no",
                        SBD_WATCHDOG_DEV=detect_watchdog_device())
    csync2_update(SYSCONFIG_SBD)
    status_done()


def init_sbd():
    """
    Configure SBD (Storage-based fencing).

    SBD can also run in diskless mode if no device
    is configured.
    """
    if not _context.sbd_device and _context.diskless_sbd:
        init_sbd_diskless()
        return
    elif not _context.sbd_device:
        # SBD device not set up by init_storage (ocfs2 template) and
        # also not passed in as command line argument - prompt user
        if _context.yes_to_all:
            warn("Not configuring SBD (%s left untouched)." % (SYSCONFIG_SBD))
            return
        status("""
Configure SBD:
  If you have shared storage, for example a SAN or iSCSI target,
  you can use it avoid split-brain scenarios by configuring SBD.
  This requires a 1 MB partition, accessible to all nodes in the
  cluster.  The device path must be persistent and consistent
  across all nodes in the cluster, so /dev/disk/by-id/* devices
  are a good choice.  Note that all data on the partition you
  specify here will be destroyed.
""")

        if not confirm("Do you wish to use SBD?"):
            warn("Not configuring SBD - STONITH will be disabled.")
            # Comment out SBD devices if present
            if os.path.isfile(SYSCONFIG_SBD):
                scfg, nmatches = sysconfig_comment_out(open(SYSCONFIG_SBD), "SBD_DEVICE")
                if nmatches > 0:
                    utils.str2file(scfg, SYSCONFIG_SBD)
                    csync2_update(SYSCONFIG_SBD)
            return

        if not check_watchdog():
            error("Watchdog device must be configured if want to use SBD!")

        if utils.is_program("sbd") is None:
            error("sbd executable not found! Cannot configure SBD.")

        configured_dev = configured_sbd_device()
        if configured_dev:
            if not confirm("SBD is already configured to use %s - overwrite?" % (configured_dev)):
                return

        dev = ""
        dev_looks_sane = False
        while not dev_looks_sane:
            dev = prompt_for_string('Path to storage device (e.g. /dev/disk/by-id/...), or "none"', r'none|\/.*', dev)
            if dev == "none":
                _context.diskless_sbd = True
                init_sbd_diskless()
                return
            if not is_block_device(dev):
                print("    That doesn't look like a block device", file=sys.stderr)
                dev = ""
            else:
                warn("All data on {} will be destroyed!".format(dev))
                if confirm('Are you sure you wish to use this device?'):
                    dev_looks_sane = True
                else:
                    dev = ""
        _context.sbd_device = dev

    if not is_block_device(_context.sbd_device):
        error("SBD device %s doesn't seem to exist" % (_context.sbd_device))

    # TODO: need to ensure watchdog is available
    # (actually, should work if watchdog unavailable, it'll just whine in the logs...)
    # TODO: what about timeouts for multipath devices?
    status_long('Initializing SBD...')
    if not invoke("sbd -d %s create" % (_context.sbd_device)):
        error("Failed to initialize SBD device %s" % (_context.sbd_device))
    status_done()

    utils.sysconfig_set(SYSCONFIG_SBD,
                        SBD_DEVICE=_context.sbd_device,
                        SBD_PACEMAKER="yes",
                        SBD_STARTMODE="always",
                        SBD_DELAY_START="no",
                        SBD_WATCHDOG_DEV=detect_watchdog_device())
    csync2_update(SYSCONFIG_SBD)


def init_cluster():
    """
    Initial cluster configuration.
    """
    init_cluster_local()

    _rc, nnodes = utils.get_stdout("crm_node -l")
    nnodes = len(nnodes.splitlines())
    if nnodes < 1:
        error("No nodes found in cluster")
    if nnodes > 1:
        error("Joined existing cluster - will not reconfigure.")

    status("Loading initial cluster configuration")

    crm_configure_load("update", """
property cib-bootstrap-options: stonith-enabled=false placement-strategy=balanced
op_defaults op-options: timeout=600 record-pending=true
rsc_defaults rsc-options: resource-stickiness=1 migration-threshold=3
""")

    if configured_sbd_device():
        if not invoke("crm configure primitive stonith-sbd stonith:external/sbd pcmk_delay_max=30s"):
            error("Can't create stonith-sbd primitive")
        if not invoke("crm configure property stonith-enabled=true"):
            error("Can't enable STONITH for SBD")
    elif _context.diskless_sbd:
        # TODO: configure stonith-watchdog-timeout correctly
        if not invoke("crm configure property stonith-enabled=true stonith-watchdog-timeout=5s"):
            error("Can't enable STONITH for diskless SBD")


def init_vgfs():
    """
    Configure cluster OCFS2 device.
    """
    dev = _context.ocfs2_device
    if not dev:
        error("vgfs stage requires -o <dev>")
    mntpoint = "/srv/clusterfs"

    if not is_block_device(dev):
        error("OCFS2 device \"{}\" does not exist".format(dev))

    # TODO: configurable mountpoint and vg name
    crm_configure_load("update", """
primitive dlm ocf:pacemaker:controld op start timeout=90 op stop timeout=100 op monitor interval=60 timeout=60
primitive clusterfs Filesystem directory=%(mntpoint)s fstype=ocfs2 device=%(dev)s \
    op monitor interval=20 timeout=40 op start timeout=60 op stop timeout=60 \
    meta target-role=Stopped
clone base-clone dlm meta interleave=true
clone c-clusterfs clusterfs meta interleave=true clone-max=8
order base-then-clusterfs inf: base-clone c-clusterfs
colocation clusterfs-with-base inf: c-clusterfs base-clone
    """ % {"mntpoint": utils.doublequote(mntpoint), "dev": utils.doublequote(dev)})

    wait_for_resource("Waiting for DLM", "dlm:0")
    wait_for_stop("Making sure filesystem is not active", "clusterfs:0")

    _rc, blkid, _err = utils.get_stdout_stderr("blkid %s" % (dev))
    if "TYPE" in blkid:
        if not confirm("Exiting filesystem found on \"{}\" - destroy?".format(dev)):
            for res in ("base-clone", "c-clusterfs"):
                invoke("crm resource stop %s" % (res))
                wait_for_stop("Waiting for resource %s to stop" % (res), res)
            invoke("crm configure delete dlm clusterfs base-group base-clone c-clusterfs base-then-clusterfs clusterfs-with-base")

    status_long("Creating OCFS2 filesystem")
    # TODO: want "-T vmstore", but this'll only fly on >2GB partition
    # Note: using undocumented '-x' switch to avoid prompting if overwriting
    # existing partition.  For the commit that introduced this, see:
    # http://oss.oracle.com/git/?p=ocfs2-tools.git;a=commit;h=8345a068479196172190f4fa287052800fa2b66f
    # TODO: if make the cluster name configurable, we need to update it here too
    if not invoke("mkfs.ocfs2 --cluster-stack pcmk --cluster-name %s -N 8 -x %s" % (_context.cluster_name, dev)):
        error("Failed to create OCFS2 filesystem on %s" % (dev))
    status_done()

    # TODO: refactor, maybe
    if not invoke("mkdir -p %s" % (mntpoint)):
        error("Can't create mountpoint %s" % (mntpoint))
    if not invoke("crm resource meta clusterfs delete target-role"):
        error("Can't start cluster filesystem clone")
    wait_for_resource("Waiting for %s to be mounted" % (mntpoint), "clusterfs:0")


def init_admin():
    # Skip this section when -y is passed
    # unless $ADMIN_IP is set
    adminaddr = _context.admin_ip
    if _context.yes_to_all and not adminaddr:
        return

    if not adminaddr:
        status("""
Configure Administration IP Address:
  Optionally configure an administration virtual IP
  address. The purpose of this IP address is to
  provide a single IP that can be used to interact
  with the cluster, rather than using the IP address
  of any specific cluster node.
""")
        if not confirm("Do you wish to configure a virtual IP address?"):
            return

        adminaddr = prompt_for_string('Virtual IP', r'([0-9]+\.){3}[0-9]+|[0-9a-fA-F]{1,4}:', "", valid_adminIP)
        if not adminaddr:
            error("Expected an IP address")

    crm_configure_load("update", 'primitive admin-ip IPaddr2 ip=%s op monitor interval=10 timeout=20' % (utils.doublequote(adminaddr)))
    wait_for_resource("Configuring virtual IP ({})".format(adminaddr), "admin-ip")


def init():
    """
    Basic init
    """
    log_start()
    init_network()


def join_ssh(seed_host):
    """
    SSH configuration for joining node.
    """
    if not seed_host:
        error("No existing IP/hostname specified (use -c option)")

    start_service("sshd.service")
    invoke("mkdir -m 700 -p /root/.ssh")

    tmpdir = tmpfiles.create_dir()
    status("Retrieving SSH keys - This may prompt for root@%s:" % (seed_host))
    invoke("scp -oStrictHostKeyChecking=no  root@%s:'/root/.ssh/id_*' %s/" % (seed_host, tmpdir))

    # This supports all SSH key types, for the case where ha-cluster-init
    # wasn't used to set up the seed node, and the user has manually
    # created, for example, DSA keys (bnc#878080)
    got_keys = 0
    for key in ("id_rsa", "id_dsa", "id_ecdsa", "id_ed25519"):
        if not os.path.exists(os.path.join(tmpdir, key)):
            continue

        if os.path.exists(os.path.join("/root/.ssh", key)):
            if not confirm("/root/.ssh/%s already exists - overwrite?" % (key)):
                continue
        invoke("mv %s* /root/.ssh/" % (os.path.join(tmpdir, key)))
        if not grep_file("/root/.ssh/authorized_keys", open("/root/.ssh/%s.pub" % (key)).read()):
            append("/root/.ssh/%s.pub" % (key), "/root/.ssh/authorized_keys")
        got_keys += 1

    if got_keys == 0:
        status("No new SSH keys installed")
    elif got_keys == 1:
        status("One new SSH key installed")
    else:
        status("%s new SSH keys installed" % (got_keys))

    # This makes sure the seed host has its own SSH keys in its own
    # authorized_keys file (again, to help with the case where the
    # user has done manual initial setup without the assistance of
    # ha-cluster-init).
    if not invoke("ssh root@%s crm cluster init ssh_remote" % (seed_host)):
        error("Can't invoke crm cluster init ssh_remote on %s" % (seed_host))


def join_csync2(seed_host):
    """
    Csync2 configuration for joining node.
    """
    if not seed_host:
        error("No existing IP/hostname specified (use -c option)")
    status_long("Configuring csync2")

    # Necessary if re-running join on a node that's been configured before.
    rmfile("/var/lib/csync2/{}.db3".format(utils.this_node()), ignore_errors=True)

    # Not automatically updating /etc/hosts - risky in the general case.
    # etc_hosts_add_me
    # local hosts_line=$(etc_hosts_get_me)
    # [ -n "$hosts_line" ] || error "No valid entry for $(hostname) in /etc/hosts - csync2 can't work"

    # If we *were* updating /etc/hosts, the next line would have "\"$hosts_line\"" as
    # the last arg (but this requires re-enabling this functionality in ha-cluster-init)
    if not invoke("ssh root@{} crm cluster init csync2_remote {}".format(seed_host, utils.this_node())):
        error("Can't invoke crm cluster init init csync2_remote on {}".format(seed_host))

    # This is necessary if syncing /etc/hosts (to ensure everyone's got the
    # same list of hosts)
    # local tmp_conf=/etc/hosts.$$
    # invoke scp root@seed_host:/etc/hosts $tmp_conf \
    #   || error "Can't retrieve /etc/hosts from seed_host"
    # install_tmp $tmp_conf /etc/hosts

    if not invoke("scp root@%s:'/etc/csync2/{csync2.cfg,key_hagroup}' /etc/csync2" % (seed_host)):
        error("Can't retrieve csync2 config from %s" % (seed_host))

    start_service("csync2.socket")

    # Sync new config out.  This goes to all hosts; csync2.cfg definitely
    # needs to go to all hosts (else hosts other than the seed and the
    # joining host won't have the joining host in their config yet).
    # Strictly, the rest of the files need only go to the new host which
    # could theoretically be effected using `csync2 -xv -P $(hostname)`,
    # but this still leaves all the other files in dirty state (becuase
    # they haven't gone to all nodes in the cluster, which means a
    # subseqent join of another node can fail its sync of corosync.conf
    # when it updates expected_votes.  Grrr...
    if not invoke('ssh root@%s "csync2 -mr / ; csync2 -fr / ; csync2 -xv"' % (seed_host)):
        print("")
        warn("csync2 run failed - some files may not be sync'd")

    status_done()


def join_ssh_merge(_cluster_node):
    status("Merging known_hosts")

    me = utils.this_node()
    hosts = [m.group(1)
             for m in re.finditer(r"^\s*host\s*([^ ;]+)\s*;", open(CSYNC2_CFG).read(), re.M)
             if m.group(1) != me]
    if not hosts:
        hosts = [_cluster_node]
        warn("Unable to extract host list from %s" % (CSYNC2_CFG))

    try:
        import parallax
    except ImportError:
        error("parallax python library is missing")

    opts = parallax.Options()
    opts.ssh_options = ['StrictHostKeyChecking=no']

    # The act of using pssh to connect to every host (without strict host key
    # checking) ensures that at least *this* host has every other host in its
    # known_hosts
    known_hosts_new = set()
    cat_cmd = "[ -e /root/.ssh/known_hosts ] && cat /root/.ssh/known_hosts || true"
    log("parallax.call {} : {}".format(hosts, cat_cmd))
    results = parallax.call(hosts, cat_cmd, opts)
    for host, result in results.items():
        if isinstance(result, parallax.Error):
            warn("Failed to get known_hosts from {}: {}".format(host, str(result)))
        else:
            if result[1]:
                known_hosts_new.update((utils.to_ascii(result[1]) or "").splitlines())
    if known_hosts_new:
        hoststxt = "\n".join(sorted(known_hosts_new))
        tmpf = utils.str2tmp(hoststxt)
        log("parallax.copy {} : {}".format(hosts, hoststxt))
        results = parallax.copy(hosts, tmpf, "/root/.ssh/known_hosts")
        for host, result in results.items():
            if isinstance(result, parallax.Error):
                warn("scp to {} failed ({}), known_hosts update may be incomplete".format(host, str(result)))


def join_cluster(seed_host):
    """
    Cluster configuration for joining node.
    """
    def get_local_nodeid():
        # for IPv6
        all_ = utils.network_v6_all()
        if not all_:
            error("Doesn't have any usable IPv6 addresses!")
        nic = sorted(all_.keys())[0]
        ipaddr = all_[nic][0].split('/')[0]
        return utils.gen_nodeid_from_ipv6(ipaddr)

    def update_nodeid(nodeid, node=None):
        # for IPv6
        if node and node != utils.this_node():
            cmd = "crm corosync set totem.nodeid %d" % nodeid
            invoke("crm cluster run '{}' {}".format(cmd, node))
        else:
            corosync.set_value("totem.nodeid", nodeid)

    invoke("curl -o /etc/corosync/authkey http://%s:5000/api/v1/corosync/authkey" % seed_host)
    invoke("curl -o /etc/corosync/corosync.conf http://%s:5000/api/v1/corosync/corosync.conf" % seed_host)

    # check if use IPv6
    ipv6_flag = False
    ipv6 = corosync.get_value("totem.ip_version")
    if ipv6 and ipv6 == "ipv6":
        ipv6_flag = True
    _context.ipv6 = ipv6_flag

    # check whether have two rings
    rrp_flag = False
    rrp = corosync.get_value("totem.rrp_mode")
    if rrp:
        rrp_flag = True

    # Need to do this if second (or subsequent) node happens to be up and
    # connected to storage while it's being repartitioned on the first node.
    probe_partitions()

    # It would be massively useful at this point if new nodes could come
    # up in standby mode, so we could query the CIB locally to see if
    # there was any further local setup that needed doing, e.g.: creating
    # mountpoints for clustered filesystems.  Unfortunately we don't have
    # that yet, so the following crawling horror takes a punt on the seed
    # node being up, then asks it for a list of mountpoints...
    """
    if _context.cluster_node:
        _rc, outp, _ = utils.get_stdout_stderr("ssh -o StrictHostKeyChecking=no root@{} 'cibadmin -Q --xpath \"//primitive\"'".format(seed_host))
        if outp:
            xml = etree.fromstring(outp)
            mountpoints = xml.xpath(' and '.join(['//primitive[@class="ocf"',
                                                  '@provider="heartbeat"',
                                                  '@type="Filesystem"]']) +
                                    '/instance_attributes/nvpair[@name="directory"]/@value')
            for m in mountpoints:
                invoke("mkdir -p {}".format(m))
    else:
        status("No existing IP/hostname specified - skipping mountpoint detection/creation")
    """
    # Bump expected_votes in corosync.conf
    # TODO(must): this is rather fragile (see related code in ha-cluster-remove)

    # If corosync.conf() doesn't exist or is empty, we will fail here. (bsc#943227)
    if not os.path.exists(corosync.conf()):
        error("{} is not readable. Please ensure that hostnames are resolvable.".format(corosync.conf()))

    # if unicast, we need to add our node to $corosync.conf()
    is_unicast = "nodelist" in open(corosync.conf()).read()
    if is_unicast:
        ringXaddr_res = []
        print("")
        for i in 0, 1:
            while True:
                ringXaddr = prompt_for_string('Address for ring{}'.format(i),
                                              r'([0-9]+\.){3}[0-9]+|[0-9a-fA-F]{1,4}:',
                                              "",
                                              valid_ucastIP,
                                              ringXaddr_res)
                if not ringXaddr:
                    error("No value for ring{}".format(i))

                _, outp = utils.get_stdout("ip addr show")
                tmp = re.findall(r' {}/[0-9]+ '.format(ringXaddr), outp, re.M)[0].strip()
                peer_ip = corosync.get_value("nodelist.node.ring{}_addr".format(i))
                # peer ring0_addr and local ring0_addr must be configured in the same network
                if peer_ip not in utils.Network(tmp):
                    print(term.render(clidisplay.error("    Peer IP {} is not in the same network: {}".format(peer_ip, tmp))))
                    continue

                ringXaddr_res.append(ringXaddr)
                break
            if not rrp_flag:
                break
        print("")
        invoke("rm -f /var/lib/heartbeat/crm/* /var/lib/pacemaker/cib/*")
        corosync.add_node_ucast(ringXaddr_res)
        csync2_update(corosync.conf())
        invoke("ssh root@{} corosync-cfgtool -R".format(seed_host))

    # if no SBD devices are configured,
    # check the existing cluster if the sbd service is enabled
    #if not configured_sbd_device() and invoke("ssh root@{} systemctl is-enabled sbd.service".format(seed_host)):
    #    _context.diskless_sbd = True

    if ipv6_flag and not is_unicast:
        # for ipv6 mcast
        # using ipv6 need nodeid configured
        local_nodeid = get_local_nodeid()
        update_nodeid(local_nodeid)

    # Initialize the cluster before adjusting quorum. This is so
    # that we can query the cluster to find out how many nodes
    # there are (so as not to adjust multiple times if a previous
    # attempt to join the cluster failed)
    init_cluster_local()

    status_long("Reloading cluster configuration")

    if ipv6_flag and not is_unicast:
        # for ipv6 mcast
        nodeid_dict = {}
        _rc, outp, _ = utils.get_stdout_stderr("crm_node -l")
        if _rc == 0:
            for line in outp.split('\n'):
                tmp = line.split()
                nodeid_dict[tmp[1]] = tmp[0]

    # apply nodelist in cluster
    if is_unicast:
        invoke("crm cluster run 'crm corosync reload'")

    def update_expected_votes():
        # get a list of nodes, excluding remote nodes
        nodelist = None
        loop_count = 0
        device_votes = 0
        nodecount = 0
        expected_votes = 0
        while True:
            rc, nodelist_text = utils.get_stdout("cibadmin -Ql --xpath '/cib/status/node_state'")
            if rc == 0:
                try:
                    nodelist_xml = etree.fromstring(nodelist_text)
                    nodelist = [n.get('uname') for n in nodelist_xml.xpath('//node_state') if n.get('remote_node') != 'true']
                    if len(nodelist) >= 2:
                        break
                except Exception:
                    break
            # timeout: 10 seconds
            if loop_count == 10:
                break
            loop_count += 1
            sleep(1)

        # Increase expected_votes
        # TODO: wait to adjust expected_votes until after cluster join,
        # so that we can ask the cluster for the current membership list
        # Have to check if a qnetd device is configured and increase
        # expected_votes in that case
        use_qdevice = 1 if corosync.get_value("quorum.device.model") == "net" else 0
        if nodelist is None:
            for v in corosync.get_values("quorum.expected_votes"):
                expected_votes = v

                # For node >= 2, expected_votes = nodecount + device_votes
                # Assume nodecount is N, for ffsplit, qdevice only has one vote
                # which means that device_votes is 1, ie:expected_votes = N + 1;
                # while for lms, qdevice has N - 1 votes, ie: expected_votes = N + (N - 1)
                # and update quorum.device.net.algorithm based on device_votes

                if corosync.get_value("quorum.device.net.algorithm") == "lms":
                    device_votes = int((expected_votes - 1) / 2)
                    nodecount = expected_votes - device_votes
                    # as nodecount will increase 1, and device_votes is nodecount - 1
                    # device_votes also increase 1
                    device_votes += 1
                elif corosync.get_value("quorum.device.net.algorithm") == "ffsplit":
                    device_votes = 1
                    nodecount = expected_votes - device_votes
                elif use_qdevice == 0:
                    device_votes = 0
                    nodecount = v

                nodecount += 1
                expected_votes = nodecount + device_votes
                corosync.set_value("quorum.expected_votes", str(expected_votes))
        else:
            nodecount = len(nodelist)
            expected_votes = 0
            # For node >= 2, expected_votes = nodecount + device_votes
            # Assume nodecount is N, for ffsplit, qdevice only has one vote
            # which means that device_votes is 1, ie:expected_votes = N + 1;
            # while for lms, qdevice has N - 1 votes, ie: expected_votes = N + (N - 1)
            if corosync.get_value("quorum.device.net.algorithm") == "ffsplit":
                device_votes = 1
            if corosync.get_value("quorum.device.net.algorithm") == "lms":
                device_votes = nodecount - 1

            expected_votes = nodecount + device_votes

            if corosync.get_value("quorum.expected_votes"):
                corosync.set_value("quorum.expected_votes", str(expected_votes))
        if use_qdevice == 0:
            corosync.set_value("quorum.two_node", 1 if expected_votes == 2 else 0)
        if use_qdevice:
            corosync.set_value("quorum.device.votes", device_votes)

        #csync2_update(corosync.conf())
        invoke("curl -F file=@%s http://%s:5000/api/v1/corosync/corosync.conf -X POST" % (corosync.conf(), seed_host))
    update_expected_votes()

    # Trigger corosync config reload to ensure expected_votes is propagated
    invoke("corosync-cfgtool -R")

    # Ditch no-quorum-policy=ignore
    _rc, outp = utils.get_stdout("crm configure show")
    if re.search('no-quorum-policy=.*ignore', outp):
        invoke("crm_attribute --attr-name no-quorum-policy --delete-attr")

    # if unicast, we need to reload the corosync configuration
    # on the other nodes
    if is_unicast:
        invoke("crm cluster run 'crm corosync reload'")

    if ipv6_flag and not is_unicast:
        # for ipv6 mcast
        # after csync2_update, all config files are same
        # but nodeid must be uniqe
        for node in list(nodeid_dict.keys()):
            if node == utils.this_node():
                continue
            update_nodeid(int(nodeid_dict[node]), node)
        update_nodeid(local_nodeid)

    status_done()


def remove_ssh():
    seed_host = _context.cluster_node
    if not seed_host:
        error("No existing IP/hostname specified (use -c option)")

    # check whether corosync has been started on local node
    if not service_is_active("corosync.service"):
        error("Cluster is not active - can't execute removing action")

    remove_get_hostname(seed_host)


def remove_get_hostname(seed_host):
    """
    Get the nodename

    Sets _context.host_status to
    0 = Disconnected
    1 = Connected
    2 = Connected through IP address
    """
    _context.connect_name = seed_host
    _context.host_status = 0

    # if node is localhost, assume connected and hostname = cluster name
    if seed_host == utils.this_node():
        _context.cluster_node = seed_host
        _context.host_status = 1
        return

    _rc, outp, _errp = utils.get_stdout_stderr("ssh root@{} \"hostname\"".format(seed_host))
    if outp:
        _context.connect_name = seed_host
        _context.cluster_node = outp.strip()
        _context.host_status = 1
    elif re.match(r'^[\[0-9].*', seed_host):
        # IP address
        warn("Could not connect to {}".format(seed_host))
        nodename = prompt_for_string('Please enter the hostname of the node to be removed', '.+' "")
        if not nodename:
            error("Not a valid hostname")

        if nodename not in xmlutil.listnodes():
            error("Specified node {} is not configured in cluster, can not remove".format(nodename))

        _rc, outp, _errp = utils.get_stdout_stderr("ssh root@{} \"hostname\"".format(nodename))
        if outp:
            _context.connect_name = seed_host
            _context.cluster_node = nodename
            _context.host_status = 1
        else:
            _context.host_status = 2
    else:
        if seed_host not in xmlutil.listnodes():
            error("Specified node {} is not configured in cluster, can not remove".format(seed_host))

        warn("Could not resolve hostname {}".format(seed_host))
        nodename = prompt_for_string('Please enter the IP address of the node to be removed (e.g: 192.168.0.1)', r'([0-9]+\.){3}[0-9]+', "")
        if not nodename:
            error("Invalid IP address")

        # try to use the IP address to connect
        _rc, outp, _errp = utils.get_stdout_stderr("ssh root@{} \"hostname\"".format(nodename))
        if outp:
            ipaddr = nodename
            nodename = outp.strip()
            if nodename != seed_host:
                warn("Specified IP address {ipaddr} is node {nodename}, not configured in cluster.".format(ipaddr=ipaddr, nodename=nodename))
                warn("Trying to remove node {}".format(seed_host))
                _context.host_status = 0
                _context.cluster_node = nodename
            else:
                _context.host_status = 2
                _context.connect_name = ipaddr
        else:
            _context.host_status = 0


def remove_node_from_cluster():
    """
    Remove node from running cluster and the corosync / pacemaker configuration.
    """
    node = _context.cluster_node

    if _context.host_status != 0:
        status("Stopping the corosync service")
        if not invoke('ssh root@{} "systemctl stop corosync"'.format(_context.connect_name)):
            error("Stopping corosync on {} failed".format(_context.connect_name))

        # delete configuration files from the node to be removed
        toremove = [SYSCONFIG_SBD, CSYNC2_CFG, corosync.conf(), CSYNC2_KEY, COROSYNC_AUTH]
        if not invoke('ssh root@{} "bash -c \\\"rm -f {} && rm -f /var/lib/heartbeat/crm/* /var/lib/pacemaker/cib/*\\\""'.format(node, " ".join(toremove))):
            error("Deleting the configuration files failed")
    else:
        # Check node status
        _, outp = utils.get_stdout("crm_node --list")
        if any("lost" in l for l in [l for l in outp.splitlines() if node in l.split()]):
            if not confirm("The node has not been cleaned up... - remove it?"):
                return

    # execute the command : crm node delete $HOSTNAME
    status("Removing the node {}".format(node))
    if not invoke("crm node delete {}".format(node)):
        error("Failed to remove {}".format(node))

    if not invoke("sed -i /{}/d {}".format(node, CSYNC2_CFG)):
        error("Removing the node {} from {} failed".format(node, CSYNC2_CFG))

    # Remove node from nodelist if this is a unicast configuration
    if "nodelist" in open(corosync.conf()).read():
        corosync.del_node(node)

    # Decrement expected_votes in corosync.conf
    use_qdevice = 1 if "net" in corosync.get_values("quorum.device.model") else 0
    for vote in corosync.get_values("quorum.expected_votes"):
        quorum = int(vote)
        new_quorum = quorum - 1
        if use_qdevice > 0:
            new_nodecount = 0
            device_votes = 0
            nodecount = 0

            if corosync.get_value("quorum.device.net.algorithm") == "lms":
                nodecount = int((quorum + 1)/2)
                new_nodecount = nodecount - 1
                device_votes = new_nodecount - 1

            elif corosync.get_value("quorum.device.net.algorithm") == "ffsplit":
                device_votes = 1
                nodecount = quorum - device_votes
                new_nodecount = nodecount - 1

            if new_nodecount == 1:
                device_votes = 0

            corosync.set_value("quorum.device.votes", device_votes)
            new_quorum = new_nodecount + device_votes

        if use_qdevice == 0:
            corosync.set_value("quorum.two_node", 1 if new_quorum == 2 else 0)
        corosync.set_value("quorum.expected_votes", str(new_quorum))

    status("Propagating configuration changes across the remaining nodes")
    csync2_update(CSYNC2_CFG)

    # Trigger corosync config reload to ensure expected_votes is propagated
    invoke("corosync-cfgtool -R")


def remove_localhost_check():
    """
    Check whether the specified node is a cluster member
    or the local node.
    """
    nodename = _context.cluster_node
    nodes = xmlutil.listnodes()
    if nodename not in nodes:
        error("Specified node {} is not configured in cluster! Unable to remove.".format(nodename))

    return nodename == utils.this_node()


def bootstrap_init(cluster_name="hacluster", ui_context=None, nic=None, ocfs2_device=None,
                   shared_device=None, sbd_device=None, diskless_sbd=False, quiet=False,
                   template=None, admin_ip=None, yes_to_all=False,
                   unicast=False, second_hb=False, ipv6=False, watchdog=None, qdevice=None, stage=None, args=None):
    """
    -i <nic>
    -o <ocfs2-device>
    -p <shared-device>
    -s <sbd-device>
    -S - configure SBD without disk
    -t <template>
    -A [<admin-ip>]
    -q - quiet
    -y - yes to all
    -u - unicast
    <stage>

    stages:
    ssh
    ssh_remote
    csync2
    csync2_remote
    corosync
    storage
    sbd
    cluster
    vgfs
    admin
    """
    global _context
    _context = Context(quiet=quiet, yes_to_all=yes_to_all, nic=nic)
    _context.cluster_name = cluster_name
    _context.ocfs2_device = ocfs2_device
    _context.shared_device = shared_device
    _context.sbd_device = sbd_device
    _context.diskless_sbd = diskless_sbd
    _context.unicast = unicast
    _context.second_hb = second_hb
    _context.ipv6 = ipv6
    _context.admin_ip = admin_ip
    _context.watchdog = watchdog
    _context.ui_context = ui_context
    _context.qdevice = qdevice

    def check_option():
        if _context.admin_ip and not valid_adminIP(_context.admin_ip):
            error("Invalid option: admin_ip")

    if stage is None:
        stage = ""

    # vgfs stage requires running cluster, everything else requires inactive cluster,
    # except ssh and csync2 (which don't care) and csync2_remote (which mustn't care,
    # just in case this breaks ha-cluster-join on another node).
    corosync_active = service_is_active("corosync.service")
    if stage in ("vgfs", "admin"):
        if not corosync_active:
            error("Cluster is inactive - can't run %s stage" % (stage))
    elif stage == "":
        if corosync_active:
            error("Cluster is currently active - can't run")
    elif stage not in ("ssh", "ssh_remote", "csync2", "csync2_remote"):
        if corosync_active:
            error("Cluster is currently active - can't run %s stage" % (stage))

    check_option()

    # Need hostname resolution to work, want NTP (but don't block ssh_remote or csync2_remote)
    if stage not in ('ssh_remote', 'csync2_remote'):
        check_tty()
        if not check_prereqs(stage):
            return
    elif stage == 'csync2_remote':
        log("args: {}".format(args))
        if len(args) != 2:
            error("Expected NODE argument to csync2_remote")
        _context.cluster_node = args[1]

    init()

    if stage != "":
        globals()["init_" + stage]()
    else:
        if watchdog is not None:
            init_watchdog()
        #init_ssh()
        #init_csync2()
        init_api_server()
        init_corosync()
        init_remote_auth()
        if template == 'ocfs2':
            if sbd_device is None or ocfs2_device is None:
                init_storage()
        #init_sbd()
        init_cluster()
        if template == 'ocfs2':
            init_vgfs()
        init_admin()

    status("Done (log saved to %s)" % (LOG_FILE))


def bootstrap_join(cluster_node=None, ui_context=None, nic=None, quiet=False, yes_to_all=False, watchdog=None, stage=None):
    """
    -c <cluster-node>
    -i <nic>
    -q - quiet
    -y - yes to all
    <stage>
    # stages:
    ssh
    csync2
    ssh_merge
    cluster
    """
    global _context
    _context = Context(quiet=quiet, yes_to_all=yes_to_all, nic=nic)
    _context.cluster_node = cluster_node
    _context.watchdog = watchdog
    _context.ui_context = ui_context

    check_tty()

    corosync_active = service_is_active("corosync.service")
    if corosync_active:
        error("Abort: Cluster is currently active. Run this command on a node joining the cluster.")

    if not check_prereqs("join"):
        return

    init()

    if stage != "":
        globals()["join_" + stage](cluster_node)
    else:
        if not yes_to_all and cluster_node is None:
            status("""Join This Node to Cluster:
  You will be asked for the IP address of an existing node, from which
  configuration will be copied.  If you have not already configured
  passwordless ssh between nodes, you will be prompted for the root
  password of the existing node.
""")
            cluster_node = prompt_for_string("IP address or hostname of existing node (e.g.: 192.168.1.1)", ".+")
            _context.cluster_node = cluster_node

        #join_ssh(cluster_node)
        join_remote_auth(cluster_node)
        #join_csync2(cluster_node)
        #join_ssh_merge(cluster_node)
        init_api_server()
        join_cluster(cluster_node)

    status("Done (log saved to %s)" % (LOG_FILE))


def join_remote_auth(node):
    if os.path.exists(PCMK_REMOTE_AUTH):
        rmfile(PCMK_REMOTE_AUTH)
    pcmk_remote_dir = os.path.dirname(PCMK_REMOTE_AUTH)
    mkdirs_owned(pcmk_remote_dir, mode=0o750, gid="haclient")
    invoke("touch {}".format(PCMK_REMOTE_AUTH))


def bootstrap_remove(cluster_node=None, ui_context=None, quiet=False, yes_to_all=False, force=False):
    """
    -c <cluster-node> - node to remove from cluster
    -q - quiet
    -y - yes to all
    -f - force removal of self
    """
    global _context
    _context = Context(quiet=quiet, yes_to_all=yes_to_all)
    _context.cluster_node = cluster_node
    _context.ui_context = ui_context

    if not yes_to_all and cluster_node is None:
        status("""Remove This Node from Cluster:
  You will be asked for the IP address or name of an existing node,
  which will be removed from the cluster. This command must be
  executed from a different node in the cluster.
""")
        cluster_node = prompt_for_string("IP address or hostname of cluster node (e.g.: 192.168.1.1)", ".+")
        _context.cluster_node = cluster_node

    init()
    remove_ssh()

    if not force and not confirm("Removing node \"{}\" from the cluster: Are you sure?".format(cluster_node)):
        return

    if remove_localhost_check():
        if not config.core.force and not force:
            error("Removing self requires --force")
        # get list of cluster nodes
        me = utils.this_node()
        nodes = xmlutil.listnodes(include_remote_nodes=False)
        othernode = next((x for x in nodes if x != me), None)
        if othernode is not None:
            # remove from other node
            cmd = "crm cluster remove{} -c {}".format(" -y" if yes_to_all else "", me)
            rc = utils.ext_cmd_nosudo("ssh{} -o StrictHostKeyChecking=no {} '{}'".format("" if yes_to_all else " -t", othernode, cmd))
            if rc != 0:
                error("Failed to remove this node from {}".format(othernode))
        else:
            # stop cluster
            if not stop_service("corosync"):
                error("Stopping corosync failed")

            # remove all trace of cluster from this node
            # delete configuration files from the node to be removed
            toremove = [
                SYSCONFIG_SBD,
                CSYNC2_CFG,
                corosync.conf(),
                CSYNC2_KEY,
                COROSYNC_AUTH,
                "/var/lib/heartbeat/crm/*",
                "/var/lib/pacemaker/cib/*"]
            if not invoke('bash -c "rm -f {}"'.format(" ".join(toremove))):
                error("Deleting the configuration files failed")
    else:
        remove_node_from_cluster()


def init_common_geo():
    """
    Tasks to do both on first and other geo nodes.
    """
    if not package_is_installed("booth"):
        error("Booth not installed - Not configurable as a geo cluster node.")


BOOTH_CFG = "/etc/booth/booth.conf"
BOOTH_AUTH = "/etc/booth/authkey"


def init_csync2_geo():
    """
    TODO: Configure csync2 for geo cluster
    That is, create a second sync group which
    syncs the geo configuration across the whole
    geo cluster.
    """


def create_booth_authkey():
    status("Create authentication key for booth")
    if os.path.exists(BOOTH_AUTH):
        rmfile(BOOTH_AUTH)
    if not invoke("booth-keygen {}".format(BOOTH_AUTH)):
        error("Failed to generate booth authkey")


def create_booth_config(arbitrator, clusters, tickets):
    status("Configure booth")

    config_template = """# The booth configuration file is "/etc/booth/booth.conf". You need to
# prepare the same booth configuration file on each arbitrator and
# each node in the cluster sites where the booth daemon can be launched.

# "transport" means which transport layer booth daemon will use.
# Currently only "UDP" is supported.
transport="UDP"
port="9929"
"""
    cfg = [config_template]
    if arbitrator is not None:
        cfg.append("arbitrator=\"{}\"".format(arbitrator))
    for s in clusters.values():
        cfg.append("site=\"{}\"".format(s))
    cfg.append("authfile=\"{}\"".format(BOOTH_AUTH))
    for t in tickets:
        cfg.append("ticket=\"{}\"\nexpire=\"600\"".format(t))
    cfg = "\n".join(cfg) + "\n"

    if os.path.exists(BOOTH_CFG):
        rmfile(BOOTH_CFG)
    utils.str2file(cfg, BOOTH_CFG)
    utils.chown(BOOTH_CFG, "hacluster", "haclient")
    os.chmod(BOOTH_CFG, 0o644)


def bootstrap_init_geo(quiet, yes_to_all, arbitrator, clusters, tickets, ui_context=None):
    """
    Configure as a geo cluster member.
    """
    global _context
    _context = Context(quiet=quiet, yes_to_all=yes_to_all)
    _context.ui_context = ui_context

    if os.path.exists(BOOTH_CFG) and not confirm("This will overwrite {} - continue?".format(BOOTH_CFG)):
        return
    if os.path.exists(BOOTH_AUTH) and not confirm("This will overwrite {} - continue?".format(BOOTH_AUTH)):
        return

    init_common_geo()

    # TODO:
    # in /etc/drbd.conf or /etc/drbd.d/global_common.conf
    # set common.startup.wfc-timeout 100
    # set common.startup.degr-wfc-timeout 120

    create_booth_authkey()
    create_booth_config(arbitrator, clusters, tickets)
    status("Sync booth configuration across cluster")
    csync2_update("/etc/booth")
    init_csync2_geo()
    geo_cib_config(clusters)


def geo_fetch_config(node):
    # TODO: clean this up
    status("Retrieving configuration - This may prompt for root@%s:" % (node))
    tmpdir = tmpfiles.create_dir()
    invoke("scp root@%s:'/etc/booth/*' %s/" % (node, tmpdir))
    try:
        if os.path.isfile("%s/authkey" % (tmpdir)):
            invoke("mv %s/authkey %s" % (tmpdir, BOOTH_AUTH))
            os.chmod(BOOTH_AUTH, 0o600)
        if os.path.isfile("%s/booth.conf" % (tmpdir)):
            invoke("mv %s/booth.conf %s" % (tmpdir, BOOTH_CFG))
            os.chmod(BOOTH_CFG, 0o644)
    except OSError as err:
        raise ValueError("Problem encountered with booth configuration from {}: {}".format(node, err))


def geo_cib_config(clusters):
    cluster_name = corosync.get_values('totem.cluster_name')[0]
    if cluster_name not in list(clusters.keys()):
        error("Local cluster name is {}, expected {}".format(cluster_name, "|".join(list(clusters.keys()))))

    status("Configure cluster resources for booth")
    crm_template = Template("""
primitive booth-ip ocf:heartbeat:IPaddr2 $iprules
primitive booth-site ocf:pacemaker:booth-site \
  meta resource-stickiness="INFINITY" \
  params config=booth op monitor interval="10s"
group g-booth booth-ip booth-site meta target-role=Stopped
""")
    iprule = 'params rule #cluster-name eq {} ip="{}"'

    crm_configure_load("update", crm_template.substitute(iprules=" ".join(iprule.format(k, v) for k, v in clusters.items())))


def bootstrap_join_geo(quiet, yes_to_all, node, clusters, ui_context=None):
    """
    Run on second cluster to add to a geo configuration.
    It fetches its booth configuration from the other node (cluster node or arbitrator).
    """
    global _context
    _context = Context(quiet=quiet, yes_to_all=yes_to_all)
    _context.ui_context = ui_context
    init_common_geo()
    check_tty()
    geo_fetch_config(node)
    status("Sync booth configuration across cluster")
    csync2_update("/etc/booth")
    geo_cib_config(clusters)


def bootstrap_arbitrator(quiet, yes_to_all, node, ui_context=None):
    """
    Configure this machine as an arbitrator.
    It fetches its booth configuration from a cluster node already in the cluster.
    """
    global _context
    _context = Context(quiet=quiet, yes_to_all=yes_to_all)
    _context.ui_context = ui_context
    init_common_geo()
    check_tty()
    geo_fetch_config(node)
    if not os.path.isfile(BOOTH_CFG):
        error("Failed to copy {} from {}".format(BOOTH_CFG, node))
    # TODO: verify that the arbitrator IP in the configuration is us?
    status("Enabling and starting the booth arbitrator service")
    start_service("booth@booth")

# EOF



# Copyright (C) 2008-2011 Dejan Muhamedagic <dmuhamedagic@suse.de>
# See COPYING for license information.

import sys
import os
import atexit
import random

from . import config
from . import options
from . import constants
from .msg import err_buf, common_err
from . import clidisplay
from . import term
from . import utils
from . import userdir

from . import ui_root
from . import ui_context


random.seed()


def load_rc(context, rcfile):
    # only load the RC file if there is no new-style user config
    if config.has_user_config():
        return

    try:
        f = open(rcfile)
    except:
        return
    save_stdin = sys.stdin
    sys.stdin = f
    while True:
        inp = utils.multi_input()
        if inp is None:
            break
        try:
            if not context.run(inp):
                raise ValueError("Error in RC file: " + rcfile)
        except ValueError as msg:
            common_err(msg)
    f.close()
    sys.stdin = save_stdin


def exit_handler():
    '''
    Write the history file. Remove tmp files.
    '''
    if options.interactive and not options.batch:
        try:
            from readline import write_history_file
            write_history_file(userdir.HISTORY_FILE)
        except:
            pass


# prefer the user set PATH
def envsetup():
    path = os.environ["PATH"].split(':')
    # always add these dirs to PATH if they exist
    libexec_dirs = ('/usr/lib64', '/usr/libexec', '/usr/lib',
                    '/usr/local/lib64', '/usr/local/libexec', '/usr/local/lib')
    pacemaker_dirs = set("{}/pacemaker".format(d) for d in libexec_dirs)
    pacemaker_dirs.add(config.path.crm_daemon_dir)
    pacemaker_dirs.add(os.path.dirname(sys.argv[0]))
    for p in pacemaker_dirs:
        if p not in path and os.path.isdir(p):
            os.environ['PATH'] = "%s:%s" % (os.environ['PATH'], p)


# three modes: interactive (no args supplied), batch (input from
# a file), half-interactive (args supplied, but not batch)
def cib_prompt():
    shadow = utils.get_cib_in_use()
    if not shadow:
        return constants.live_cib_prompt
    if constants.tmp_cib:
        return constants.tmp_cib_prompt
    return shadow


def make_option_parser():
    from optparse import OptionParser
    parser = OptionParser(usage="""%prog [-h|--help] [OPTIONS] [SUBCOMMAND ARGS...]
or %prog help SUBCOMMAND

For a list of available subcommands, use %prog help.

Use %prog without arguments for an interactive session.
Call a subcommand directly for a "single-shot" use.
Call %prog with a level name as argument to start an interactive
session from that level.

See the crm(8) man page or call %prog help for more details.""",
                          version="%prog " + config.CRM_VERSION)
    parser.disable_interspersed_args()
    parser.add_option("-f", "--file", dest="filename", metavar="FILE",
                      help="Load commands from the given file. If a dash (-) " +
                      "is used in place of a file name, crm will read commands " +
                      "from the shell standard input (stdin).")
    parser.add_option("-c", "--cib", dest="cib", metavar="CIB",
                      help="Start the session using the given shadow CIB file. " +
                      "Equivalent to `cib use <CIB>`.")
    parser.add_option("-D", "--display", dest="display", metavar="OUTPUT_TYPE",
                      help="Choose one of the output options: plain, color-always, color, or uppercase. " +
                      "The default is color if the terminal emulation supports colors, " +
                      "else plain.")
    parser.add_option("-F", "--force", action="store_true", default=False, dest="force",
                      help="Make crm proceed with applying changes where it would normally " +
                      "ask the user to confirm before proceeding. This option is mainly useful " +
                      "in scripts, and should be used with care.")
    parser.add_option("-n", "--no", action="store_true", default=False, dest="ask_no",
                      help="Automatically answer no when prompted")
    parser.add_option("-w", "--wait", action="store_true", default=False, dest="wait",
                      help="Make crm wait for the cluster transition to finish " +
                      "(for the changes to take effect) after each processed line.")
    parser.add_option("-H", "--history", dest="history", metavar="DIR|FILE|SESSION",
                      help="A directory or file containing a cluster report to load " +
                      "into history, or the name of a previously saved history session.")
    parser.add_option("-d", "--debug", action="store_true", default=False, dest="debug",
                      help="Print verbose debugging information.")
    parser.add_option("-R", "--regression-tests", action="store_true", default=False,
                      dest="regression_tests",
                      help="Enables extra verbose trace logging used by the regression " +
                      "tests. Logs all external calls made by crmsh.")
    parser.add_option("--scriptdir", dest="scriptdir", metavar="DIR",
                      help="Extra directory where crm looks for cluster scripts, or a list " +
                      "of directories separated by semi-colons (e.g. /dir1;/dir2;etc.).")
    parser.add_option("-X", dest="profile", metavar="PROFILE",
                      help="Collect profiling data and save in PROFILE.")
    parser.add_option("-o", "--opt", action="append", type="string", metavar="OPTION=VALUE",
                      help="Set crmsh option temporarily. If the options are saved using" +
                      "+options save+ then the value passed here will also be saved." +
                      "Multiple options can be set by using +-o+ multiple times.")
    return parser


option_parser = make_option_parser()


def usage(rc):
    option_parser.print_usage(file=(sys.stderr if rc != 0 else sys.stdout))
    sys.exit(rc)


def set_interactive():
    '''Set the interactive option only if we're on a tty.'''
    if utils.can_ask():
        options.interactive = True


def compatibility_setup():
    if not utils.is_pcmk_118():
        del constants.attr_defaults["node"]
        constants.cib_no_section_rc = 22


def add_quotes(args):
    '''
    Add quotes if there's whitespace in one of the
    arguments; so that the user doesn't need to protect the
    quotes.

    If there are two kinds of quotes which actually _survive_
    the getopt, then we're _probably_ screwed.

    At any rate, stuff like ... '..."..."'
    as well as '...\'...\''  do work.
    '''
    l = []
    for s in args:
        if config.core.add_quotes and ' ' in s:
            q = '"' in s and "'" or '"'
            if q not in s:
                s = "%s%s%s" % (q, s, q)
        l.append(s)
    return l


def handle_noninteractive_use(context, user_args):
    """
    returns: either a status code of 0 or 1, or
    None to indicate that nothing was done here.
    """
    if options.shadow:
        if not context.run("cib use " + options.shadow):
            return 1

    # this special case is silly, but we have to keep it to
    # preserve the backward compatibility
    if len(user_args) == 1 and user_args[0].startswith("conf"):
        if not context.run("configure"):
            return 1
    elif len(user_args) > 0:
        # we're not sure yet whether it's an interactive session or not
        # (single-shot commands aren't)
        err_buf.reset_lineno()
        options.interactive = False

        l = add_quotes(user_args)
        if context.run(' '.join(l)):
            # if the user entered a level, then just continue
            if not context.previous_level():
                return 0
            set_interactive()
            if options.interactive:
                err_buf.reset_lineno(-1)
        else:
            return 1
    return None


def render_prompt(context):
    rendered_prompt = constants.prompt
    if options.interactive and not options.batch:
        # TODO: fix how color interacts with readline,
        # seems the color prompt messes it up
        promptstr = "crm(%s/%s)%s# " % (cib_prompt(), utils.this_node(), context.prompt())
        constants.prompt = promptstr
        if clidisplay.colors_enabled():
            rendered_prompt = term.render(clidisplay.prompt(promptstr))
        else:
            rendered_prompt = promptstr
    return rendered_prompt


def setup_context(context):
    if options.input_file and options.input_file != "-":
        try:
            sys.stdin = open(options.input_file)
        except IOError as msg:
            common_err(msg)
            usage(2)

    if options.interactive and not options.batch:
        context.setup_readline()


def main_input_loop(context, user_args):
    """
    Main input loop for crmsh. Parses input
    line by line.
    """
    compatibility_setup()
    rc = handle_noninteractive_use(context, user_args)
    if rc is not None:
        return rc

    setup_context(context)

    rc = 0
    while True:
        try:
            inp = utils.multi_input(render_prompt(context))
            if inp is None:
                if options.interactive:
                    rc = 0
                context.quit(rc)
            try:
                if not context.run(inp):
                    rc = 1
            except ValueError as msg:
                rc = 1
                common_err(msg)
        except KeyboardInterrupt:
            if options.interactive and not options.batch:
                print("Ctrl-C, leaving")
            context.quit(1)
    return rc


def compgen():
    args = sys.argv[2:]
    if len(args) < 2:
        return

    options.shell_completion = True

    # point = int(args[0])
    line = args[1]

    # remove [*]crm from commandline
    idx = line.find('crm')
    if idx >= 0:
        line = line[idx+3:].lstrip()

    options.interactive = False
    ui = ui_root.Root()
    context = ui_context.Context(ui)
    last_word = line.rsplit(' ', 1)
    if len(last_word) > 1 and ':' in last_word[1]:
        idx = last_word[1].rfind(':')
        for w in context.complete(line):
            print(w[idx+1:])
    else:
        for w in context.complete(line):
            print(w)


def parse_options():
    opts, args = option_parser.parse_args()
    config.core.debug = "yes" if opts.debug else config.core.debug
    options.profile = opts.profile or options.profile
    options.regression_tests = opts.regression_tests or options.regression_tests
    config.color.style = opts.display or config.color.style
    config.core.force = opts.force or config.core.force
    if opts.filename:
        err_buf.reset_lineno()
        options.input_file, options.batch, options.interactive = opts.filename, True, False
    options.history = opts.history or options.history
    config.core.wait = opts.wait or config.core.wait
    options.shadow = opts.cib or options.shadow
    options.scriptdir = opts.scriptdir or options.scriptdir
    options.ask_no = opts.ask_no
    for opt in opts.opt or []:
        try:
            k, v = opt.split('=')
            s, n = k.split('.')
            config.set_option(s, n, v)
        except ValueError as e:
            raise ValueError("Expected -o <section>.<name>=<value>: %s" % (e))
    return args


def profile_run(context, user_args):
    import cProfile
    cProfile.runctx('main_input_loop(context, user_args)',
                    globals(),
                    {'context': context, 'user_args': user_args},
                    filename=options.profile)
    # print how to use the profile file, but don't disturb
    # the regression tests
    if not options.regression_tests:
        stats_cmd = "; ".join(['import pstats',
                               's = pstats.Stats("%s")' % options.profile,
                               's.sort_stats("cumulative").print_stats()'])
        print("python -c '%s' | less" % (stats_cmd))
    return 0


def run():
    try:
        if len(sys.argv) >= 2 and sys.argv[1] == '--compgen':
            compgen()
            return 0
        envsetup()
        userdir.mv_user_files()

        ui = ui_root.Root()
        context = ui_context.Context(ui)

        load_rc(context, userdir.RC_FILE)
        atexit.register(exit_handler)
        options.interactive = utils.can_ask()
        if not options.interactive:
            err_buf.reset_lineno()
            options.batch = True
        user_args = parse_options()
        term.init()
        if options.profile:
            return profile_run(context, user_args)
        else:
            return main_input_loop(context, user_args)
    except KeyboardInterrupt:
        print("Ctrl-C, leaving")
        sys.exit(1)
    except ValueError as e:
        if config.core.debug:
            import traceback
            traceback.print_exc()
            sys.stdout.flush()
        common_err(str(e))

# vim:ts=4:sw=4:et:



# Copyright (C) 2008-2011 Dejan Muhamedagic <dmuhamedagic@suse.de>
# Copyright (C) 2013 Kristoffer Gronlund <kgronlund@suse.com>
# See COPYING for license information.

import re
from . import command
from . import completers as compl
from . import utils
from . import ra
from . import constants
from . import options


def complete_class_provider_type(args):
    '''
    When need to show the agent supported by cluster,
    grouped by different classes or providers, Tab by Tab,
    instead of completing all the agents at one time.

    First completion:  agent classes
    Second completion: providers of ocf or types of other classes
    Last completion:   types of ocf agents
    '''
    c = args[-1]
    classes = ra.ra_classes()
    if not c:
        # First completion:  agent classes
        return [l+":" for l in classes]

    ret = set([])
    providers = ra.ra_providers_all('ocf')
    if c == 'ocf:' or re.search(r'ocf:[^:]+$', c):
        # Second completion: providers of ocf
        for p in providers:
            if p == '.isolation':
                continue
            ret.add('%s:%s:' % ('ocf', p))
        return list(ret)

    if re.search(r'ocf:.+:', c):
        # Last completion:   types of ocf agents
        for p in providers:
            if p == '.isolation':
                continue
            types = ra.ra_types('ocf', p)
            for t in types:
                ret.add('%s:%s:%s' % ('ocf', p, t))
        return list(ret)

    if c.endswith(':') and c.strip(':') in classes:
        # Second completion: types of other classes
        types = ra.ra_types(c)
        for t in types:
            ret.add('%s:%s' % (c.strip(':'), t))
        return list(ret)

    if re.search(r'.+:.+', c):
        # Second completion: types of other classes
        types = ra.ra_types(c.split(':')[0])
        for t in types:
            ret.add('%s:%s' % (c.split(':')[0], t))
        return list(ret)

    # First completion:  agent classes
    return [l+":" for l in classes]

class RA(command.UI):
    '''
    CIB shadow management class
    '''
    name = "ra"
    provider_classes = ["ocf"]

    def do_classes(self, context):
        "usage: classes"
        for c in ra.ra_classes():
            if c in self.provider_classes:
                providers = ra.ra_providers_all(c)
                if providers:
                    print("%s / %s" % (c, ' '.join(providers)))
            else:
                print("%s" % c)

    @command.skill_level('administrator')
    def do_providers(self, context, ra_type, ra_class="ocf"):
        "usage: providers <ra> [<class>]"
        print(' '.join(ra.ra_providers(ra_type, ra_class)))

    @command.skill_level('administrator')
    @command.completers(compl.call(ra.ra_classes), lambda args: ra.ra_providers_all(args[1]))
    def do_list(self, context, class_, provider_=None):
        "usage: list <class> [<provider>]"
        if class_ not in ra.ra_classes():
            context.fatal_error("class %s does not exist" % class_)
        if provider_ and provider_ not in ra.ra_providers_all(class_):
            context.fatal_error("there is no provider %s for class %s" % (provider_, class_))
        types = ra.ra_types(class_, provider_)
        if options.regression_tests:
            for t in types:
                print(t)
        else:
            utils.multicolumn(types)

    @command.skill_level('administrator')
    @command.alias('meta')
    @command.completers(complete_class_provider_type)
    def do_info(self, context, *args):
        "usage: info [<class>:[<provider>:]]<type>"
        if len(args) == 0:
            context.fatal_error("Expected [<class>:[<provider>:]]<type>")
        elif len(args) > 1:  # obsolete syntax
            if len(args) < 3:
                ra_type, ra_class, ra_provider = args[0], args[1], "heartbeat"
            else:
                ra_type, ra_class, ra_provider = args[0], args[1], args[2]
        elif args[0] in constants.meta_progs:
            ra_class, ra_provider, ra_type = args[0], None, None
        elif args[0] in constants.meta_progs_20:
            ra_class, ra_provider, ra_type = args[0], None, None
        else:
            ra_class, ra_provider, ra_type = ra.disambiguate_ra_type(args[0])
        agent = ra.RAInfo(ra_class, ra_type, ra_provider)
        if agent.mk_ra_node() is None:
            return False
        try:
            utils.page_string(agent.meta_pretty())
        except Exception as msg:
            context.fatal_error(msg)

    @command.skill_level('administrator')
    def do_validate(self, context, agentname, *params):
        "usage: validate [<class>:[<provider>:]]<type> [<key>=<value> ...]"
        rc, _ = ra.validate_agent(agentname, dict([param.split('=', 1) for param in params]), log=True)
        return rc == 0



# Copyright (C) 2011 Dejan Muhamedagic <dmuhamedagic@suse.de>
# See COPYING for license information.
#
# log pattern specification
#
# patterns are grouped one of several classes:
#  - resource: pertaining to a resource
#  - node: pertaining to a node
#  - quorum: quorum changes
#  - events: other interesting events (core dumps, etc)
#
# paterns are grouped based on a detail level
# detail level 0 is the lowest, i.e. should match the least
# number of relevant messages

# NB:
# %% stands for whatever user input we get, for instance a
# resource name or node name or just some regular expression
# in optimal case, it should be surrounded by literals
#
# [Note that resources may contain clone numbers!]

from . import utils

__all__ = ('patterns',)

_patterns_old = {
    "resource": (
        (  # detail 0
            "lrmd.*%% (?:start|stop|promote|demote|migrate)",
            "lrmd.*RA output: .%%:.*:stderr",
            "lrmd.*WARN: Managed %%:.*exited",
            "lrmd.*WARN: .* %% .*timed out$",
            "crmd.*LRM operation %%_(?:start|stop|promote|demote|migrate)_.*confirmed=true",
            "crmd.*LRM operation %%_.*Timed Out",
            "[(]%%[)][[]",
        ),
        (  # detail 1
            "lrmd.*%% (?:probe|notify)",
            "lrmd.*Managed %%:.*exited",
        ),
    ),
    "node": (
        (  # detail 0
            " %% .*Corosync.Cluster.Engine",
            " %% .*Executive.Service.RELEASE",
            " %% .*Requesting.shutdown",
            " %% .*Shutdown.complete",
            " %% .*Configuration.validated..Starting.heartbeat",
            "pengine.*Scheduling Node %% for STONITH",
            "crmd.* of %% failed",
            "stonith-ng.*host '%%'",
            "Exec.*on %% ",
            "Node %% will be fenced",
            "stonith-ng.*for %% timed",
            "stonith-ng.*can not fence %%:",
            "stonithd.*Succeeded.*node %%:",
            "(?:lost|memb): %% ",
            "crmd.*(?:NEW|LOST):.* %% ",
            "Node return implies stonith of %% ",
        ),
        (  # detail 1
        ),
    ),
    "quorum": (
        (  # detail 0
            "crmd.*Updating.quorum.status",
            "crmd.*quorum.(?:lost|ac?quir)",
        ),
        (  # detail 1
        ),
    ),
    "events": (
        (  # detail 0
            "CRIT:",
            "ERROR:",
        ),
        (  # detail 1
            "WARN:",
        ),
    ),
}

_patterns_118 = {
    "resource": (
        (  # detail 0
            "crmd.*Initiating.*%%_(?:start|stop|promote|demote|migrate)_",
            "lrmd.*operation_finished: %%_",
            "lrmd.*executing - rsc:%% action:(?:start|stop|promote|demote|migrate)",
            "lrmd.*finished - rsc:%% action:(?:start|stop|promote|demote|migrate)",

            "crmd.*LRM operation %%_(?:start|stop|promote|demote|migrate)_.*confirmed=true",
            "crmd.*LRM operation %%_.*Timed Out",
            "[(]%%[)][[]",
        ),
        (  # detail 1
            "crmd.*Initiating.*%%_(?:monitor_0|notify)",
            "lrmd.*executing - rsc:%% action:(?:monitor_0|notify)",
            "lrmd.*finished - rsc:%% action:(?:monitor_0|notify)",
        ),
    ),
    "node": (
        (  # detail 0
            " %% .*Corosync.Cluster.Engine",
            " %% .*Executive.Service.RELEASE",
            " %% .*crm_shutdown:.Requesting.shutdown",
            " %% .*pcmk_shutdown:.Shutdown.complete",
            " %% .*Configuration.validated..Starting.heartbeat",
            "pengine.*Scheduling Node %% for STONITH",
            "pengine.*Node %% will be fenced",
            "crmd.*for %% failed",
            "stonith-ng.*host '%%'",
            "Exec.*on %% ",
            "Node %% will be fenced",
            "stonith-ng.*on %% for.*timed out",
            "stonith-ng.*can not fence %%:",
            "stonithd.*Succeeded.*node %%:",
            "(?:lost|memb): %% ",
            "crmd.*(?:NEW|LOST|new|lost):.* %% ",
            "Node return implies stonith of %% ",
        ),
        (  # detail 1
        ),
    ),
    "quorum": (
        (  # detail 0
            "crmd.*Updating.(quorum).status",
            r"crmd.*quorum.(?:lost|ac?quir[^\s]*)",
        ),
        (  # detail 1
        ),
    ),
    "events": (
        (  # detail 0
            "(CRIT|crit|ERROR|error|UNCLEAN|unclean):",
        ),
        (  # detail 1
            "(WARN|warning):",
        ),
    ),
}


def patterns(cib_f=None):
    is118 = utils.is_pcmk_118(cib_f=cib_f)
    if is118:
        return _patterns_118
    else:
        return _patterns_old



# Modified pssh
# Copyright (c) 2011, Dejan Muhamedagic
# Copyright (c) 2009, Andrew McNabb
# Copyright (c) 2003-2008, Brent N. Chun

"""Parallel ssh to the set of nodes in hosts.txt.

For each node, this essentially does an "ssh host -l user prog [arg0] [arg1]
...". The -o option can be used to store stdout from each remote node in a
directory.  Each output file in that directory will be named by the
corresponding remote node's hostname or IP address.
"""

import os
import glob

from parallax.manager import Manager, FatalError
from parallax.task import Task
from parallax import Options

from .msg import common_err, common_debug, common_warn
from . import config


_DEFAULT_TIMEOUT = 60
_EC_LOGROT = 120


def parse_args(outdir, errdir, t=_DEFAULT_TIMEOUT):
    '''
    Parse the given commandline arguments.
    '''
    opts = Options()
    opts.timeout = int(t)
    opts.quiet = True
    opts.inline = False
    opts.outdir = outdir
    opts.errdir = errdir
    return opts


def get_output(odir, host):
    '''
    Looks for the output returned by the given host.
    This is somewhat problematic, since it is possible that
    different hosts can have similar hostnames. For example naming
    hosts "host.1" and "host.2" will confuse this code.
    '''
    l = []
    for fname in ["%s/%s" % (odir, host)] + glob.glob("%s/%s.[0-9]*" % (odir, host)):
        try:
            if os.path.isfile(fname):
                l += open(fname).readlines()
        except IOError:
            continue
    return l


def show_output(odir, hosts, desc):
    '''
    Display output from hosts. See get_output for caveats.
    '''
    for host in hosts:
        out_l = get_output(odir, host)
        if out_l:
            print("%s %s:" % (host, desc))
            print(''.join(out_l))


def do_pssh(l, opts):
    '''
    Adapted from psshlib. Perform command across list of hosts.
    l = [(host, command), ...]
    '''
    if opts.outdir and not os.path.exists(opts.outdir):
        os.makedirs(opts.outdir)
    if opts.errdir and not os.path.exists(opts.errdir):
        os.makedirs(opts.errdir)
    manager = Manager(opts)
    user = ""
    port = ""
    hosts = []
    for host, cmdline in l:
        cmd = ['ssh', host,
               '-o', 'PasswordAuthentication=no',
               '-o', 'SendEnv=PARALLAX_NODENUM',
               '-o', 'StrictHostKeyChecking=no']
        if hasattr(opts, 'options'):
            for opt in opts.options:
                cmd += ['-o', opt]
        if user:
            cmd += ['-l', user]
        if port:
            cmd += ['-p', port]
        if hasattr(opts, 'extra'):
            cmd.extend(opts.extra)
        if cmdline:
            cmd.append(cmdline)
        hosts.append(host)
        t = Task(host, port, user, cmd,
                 stdin=opts.input_stream,
                 verbose=opts.verbose,
                 quiet=opts.quiet,
                 print_out=opts.print_out,
                 inline=opts.inline,
                 inline_stdout=opts.inline_stdout,
                 default_user=opts.default_user)
        manager.add_task(t)
    try:
        return manager.run()  # returns a list of exit codes
    except FatalError:
        common_err("SSH to nodes failed")
        show_output(opts.errdir, hosts, "stderr")
        return False


def examine_outcome(l, opts, statuses):
    '''
    A custom function to show stderr in case there were issues.
    Not suited for callers who want better control of output or
    per-host processing.
    '''
    hosts = [x[0] for x in l]
    if min(statuses) < 0:
        # At least one process was killed.
        common_err("ssh process was killed")
        show_output(opts.errdir, hosts, "stderr")
        return False
    # The any builtin was introduced in Python 2.5 (so we can't use it yet):
    # elif any(x==255 for x in statuses):
    for status in statuses:
        if status == 255:
            common_warn("ssh processes failed")
            show_output(opts.errdir, hosts, "stderr")
            return False
    for status in statuses:
        if status not in (0, _EC_LOGROT):
            common_warn("some ssh processes failed")
            show_output(opts.errdir, hosts, "stderr")
            return False
    return True


def next_loglines(a, outdir, errdir, from_time):
    '''
    pssh to nodes to collect new logs.
    '''
    l = []
    for node, rptlog, logfile, nextpos in a:
        common_debug("updating %s from %s (pos %d)" %
                     (logfile, node, nextpos))
        if logfile.startswith("/tmp") and logfile.endswith("/journal.log"):
            cmdline = "/usr/bin/journalctl -o short-iso --since '%s' --no-pager" % (from_time)
        else:
            cmdline = "perl -e 'exit(%d) if (stat(\"%s\"))[7]<%d' && tail -c +%d %s" % (
                _EC_LOGROT, logfile, nextpos-1, nextpos, logfile)
        opts = parse_args(outdir, errdir)
        l.append([node, cmdline])
    statuses = do_pssh(l, opts)
    if statuses:
        return examine_outcome(l, opts, statuses)
    else:
        return False


def next_peinputs(node_pe_l, outdir, errdir):
    '''
    pssh to nodes to collect new logs.
    '''
    pe_dir = config.path.pe_state_dir
    vardir = os.path.dirname(pe_dir)
    l = []
    for node, pe_l in node_pe_l:
        red_pe_l = [os.path.join("pengine", os.path.basename(x)) for x in pe_l]
        cmdline = "tar -C %s -chf - %s" % (vardir, ' '.join(red_pe_l))
        common_debug("getting new PE inputs %s from %s" % (red_pe_l, node))
        opts = parse_args(outdir, errdir)
        l.append([node, cmdline])
    if not l:
        # is this a failure?
        return True
    statuses = do_pssh(l, opts)
    if statuses:
        return examine_outcome(l, opts, statuses)
    else:
        return False


def do_pssh_cmd(cmd, node_l, outdir, errdir, timeout=20000):
    '''
    pssh to nodes and run cmd.
    '''
    l = []
    for node in node_l:
        l.append([node, cmd])
    if not l:
        return True
    opts = parse_args(outdir, errdir, t=int(timeout // 1000))
    return do_pssh(l, opts)

# vim:ts=4:sw=4:et:



# Copyright (C) 2008-2011 Dejan Muhamedagic <dmuhamedagic@suse.de>
# Copyright (C) 2013 Kristoffer Gronlund <kgronlund@suse.com>
# See COPYING for license information.

from . import command
from . import completers as compl
from . import utils
from . import ui_utils
from . import constants
from .cibstatus import cib_status


_status_node_list = compl.call(cib_status.status_node_list)


class CibStatusUI(command.UI):
    '''
    The CIB status section management user interface class
    '''
    name = "cibstatus"

    @command.skill_level('expert')
    def do_load(self, context, org):
        "usage: load {<file>|shadow:<cib>|live}"
        return cib_status.load(org)

    @command.skill_level('expert')
    def do_save(self, context, dest=None):
        "usage: save [<file>|shadow:<cib>]"
        return cib_status.save(dest)

    @command.skill_level('administrator')
    def do_origin(self, context):
        "usage: origin"
        state = cib_status.modified and " (modified)" or ""
        print("%s%s" % (cib_status.origin, state))

    @command.skill_level('administrator')
    @command.completers(compl.choice(['changed']))
    def do_show(self, context, changed=""):
        "usage: show [changed]"
        if changed:
            if changed != "changed":
                context.fatal_error("Expected 'changed', got '%s'" % (changed))
            return cib_status.list_changes()
        return cib_status.show()

    @command.skill_level('administrator')
    @command.completers(compl.booleans)
    def do_quorum(self, context, opt):
        "usage: quorum <bool>"
        if not utils.verify_boolean(opt):
            context.fatal_error("%s: bad boolean option" % opt)
        return cib_status.set_quorum(utils.is_boolean_true(opt))

    @command.skill_level('expert')
    @command.completers(_status_node_list, compl.choice(constants.node_states))
    def do_node(self, context, node, state):
        "usage: node <node> {online|offline|unclean}"
        return cib_status.edit_node(node, state)

    @command.skill_level('expert')
    @command.completers(compl.null, compl.choice(list(cib_status.ticket_ops.keys())))
    def do_ticket(self, context, ticket, subcmd):
        "usage: ticket <ticket> {grant|revoke|activate|standby}"
        return cib_status.edit_ticket(ticket, subcmd)

    @command.skill_level('expert')
    @command.completers(compl.choice(constants.ra_operations),
                        compl.call(cib_status.status_rsc_list),
                        compl.choice(list(constants.lrm_exit_codes.keys())),
                        compl.choice(list(constants.lrm_status_codes.keys())),
                        compl.choice(constants.node_states))
    def do_op(self, context, op, rsc, rc, op_status=None, node=''):
        "usage: op <operation> <resource> <exit_code> [<op_status>] [<node>]"
        if rc in constants.lrm_exit_codes:
            num_rc = constants.lrm_exit_codes[rc]
        else:
            num_rc = rc
        if not num_rc.isdigit():
            context.fatal_error("Invalid exit code '%s'" % num_rc)
        num_op_status = op_status
        if op_status:
            if op_status in constants.lrm_status_codes:
                num_op_status = constants.lrm_status_codes[op_status]
            if not num_op_status.isdigit():
                context.fatal_error("Invalid operation status '%s'" % num_op_status)
        return cib_status.edit_op(op, rsc, num_rc, num_op_status, node)

    @command.skill_level('administrator')
    @command.completers(compl.choice(['nograph']))
    def do_run(self, context, *args):
        "usage: run [nograph] [v...] [scores] [utilization]"
        return ui_utils.ptestlike(cib_status.run, '', context.get_command_name(), args)

    @command.skill_level('administrator')
    @command.completers(compl.choice(['nograph']))
    def do_simulate(self, context, *args):
        "usage: simulate [nograph] [v...] [scores] [utilization]"
        return ui_utils.ptestlike(cib_status.simulate, '', context.get_command_name(), args)



# Copyright (C) 2013 Kristoffer Gronlund <kgronlund@suse.com>
# See COPYING for license information.

# Helper completers

from . import xmlutil


def choice(lst):
    '''
    Static completion from a list
    '''
    def completer(args):
        return lst
    return completer


null = choice([])
attr_id = choice(["id="])

def call(fn, *fnargs):
    '''
    Call the given function with the given arguments.
    The function has to return a list of completions.
    '''
    def completer(args):
        return fn(*fnargs)
    return completer


def join(*fns):
    '''
    Combine the output of several completers
    into a single completer.
    '''
    def completer(args):
        ret = []
        for fn in fns:
            ret += list(fn(args))
        return ret
    return completer


booleans = choice(['yes', 'no', 'true', 'false', 'on', 'off'])


def resources(args):
    cib_el = xmlutil.resources_xml()
    if cib_el is None:
        return []
    nodes = xmlutil.get_interesting_nodes(cib_el, [])
    res = [x.get("id") for x in nodes if xmlutil.is_resource(x)]
    if args and args[0] in ['promote', 'demote']:
        return list(filter(xmlutil.RscState().is_ms, res))
    if args and args[0] == "started":
        return list(filter(xmlutil.RscState().is_running, res))
    if args and args[0] == "stopped":
        for item in filter(xmlutil.RscState().is_running, res):
            res.remove(item)
    return res


def resources_started(args):
    return resources(["started"])


def resources_stopped(args):
    return resources(["stopped"])


def primitives(args):
    cib_el = xmlutil.resources_xml()
    if cib_el is None:
        return []
    nodes = xmlutil.get_interesting_nodes(cib_el, [])
    return [x.get("id") for x in nodes if xmlutil.is_primitive(x)]


nodes = call(xmlutil.listnodes)

shadows = call(xmlutil.listshadows)

status_option = """full bynode inactive ops timing failcounts
                   verbose quiet xml simple tickets noheaders
                   detail brief""".split()



# Copyright (C) 2008-2011 Dejan Muhamedagic <dmuhamedagic@suse.de>
# See COPYING for license information.

import sys
from . import config
from . import clidisplay
from . import options

ERR_STREAM = sys.stderr


class ErrorBuffer(object):
    '''
    Show error messages either immediately or buffered.
    '''
    def __init__(self):
        try:
            from . import term
            self._term = term
        except:
            self._term = None
        self.msg_list = []
        self.mode = "immediate"
        self.lineno = -1
        self.__save_lineno = -1
        self.written = {}

    def buffer(self):
        self.mode = "keep"

    def release(self):
        if self.msg_list:
            if ERR_STREAM:
                print('\n'.join(self.msg_list), file=ERR_STREAM)
            else:
                print('\n'.join(self.msg_list))
            if not options.batch:
                try:
                    input("Press enter to continue... ")
                except EOFError:
                    pass
            self.msg_list = []
        self.mode = "immediate"

    def writemsg(self, msg, to=None):
        if to is None:
            to = ERR_STREAM
        if msg.endswith('\n'):
            msg = msg[:-1]
        if self.mode == "immediate":
            if options.regression_tests or not to:
                print(msg)
            else:
                print(msg, file=to)
        else:
            self.msg_list.append(msg)

    def reset_lineno(self, to=0):
        self.lineno = to

    def incr_lineno(self):
        if self.lineno >= 0:
            self.lineno += 1

    def start_tmp_lineno(self):
        self.__save_lineno = self.lineno
        self.reset_lineno()

    def stop_tmp_lineno(self):
        self.lineno = self.__save_lineno

    def add_lineno(self, s):
        if self.lineno > 0:
            return "%d: %s" % (self.lineno, s)
        else:
            return s

    def _prefix(self, pfx, s, to=None):
        self.writemsg(self._render("%s: %s" % (pfx, self.add_lineno(s))), to=to)

    def ok(self, s):
        self._prefix(clidisplay.ok("OK"), s, to=sys.stdout)

    def error(self, s):
        self._prefix(clidisplay.error("ERROR"), s)

    def warning(self, s):
        self._prefix(clidisplay.warn("WARNING"), s)

    def one_warning(self, s):
        if s not in self.written:
            self.written[s] = 1
            self.writemsg(self._render(clidisplay.warn("WARNING")) + ": %s" %
                          self.add_lineno(s))

    def info(self, s):
        self._prefix(clidisplay.info("INFO"), s)

    def debug(self, s):
        if config.core.debug:
            self._prefix("DEBUG", s)

    def _render(self, s):
        'Render for TERM.'
        if self._term:
            return self._term.render(s)
        return s


def common_error(s):
    err_buf.error(s)


def common_err(s):
    err_buf.error(s)


def common_warning(s):
    err_buf.warning(s)


def common_warn(s):
    err_buf.warning(s)


def warn_once(s):
    err_buf.one_warning(s)


def common_info(s):
    err_buf.info(s)


def common_debug(s):
    err_buf.debug(s)


def no_prog_err(name):
    err_buf.error("%s not available, check your installation" % name)


def unsupported_err(name):
    err_buf.error("%s is not supported" % name)


def missing_obj_err(node):
    err_buf.error("object %s:%s missing (shouldn't have happened)" %
                  (node.tag, node.get("id")))


def constraint_norefobj_err(constraint_id, obj_id):
    err_buf.error("constraint %s references a resource %s which doesn't exist" %
                  (constraint_id, obj_id))


def no_object_err(name):
    err_buf.error("object %s does not exist" % name)


def invalid_id_err(obj_id):
    err_buf.error("%s: invalid object id" % obj_id)


def id_used_err(node_id):
    err_buf.error("%s: id is already in use" % node_id)


def syntax_err(s, token='', context='', msg=''):
    err = "syntax"
    if context:
        err += " in "
        err += context
    if msg:
        err += ": %s" % (msg)
    if isinstance(s, str):
        err += " parsing '%s'" % (s)
    elif token:
        err += " near <%s> parsing '%s'" % (token, ' '.join(s))
    else:
        err += " parsing '%s'" % (' '.join(s))
    err_buf.error(err)


def bad_usage(cmd, args, msg=None):
    if not msg:
        err_buf.error("Bad usage: '%s %s'" % (cmd, args))
    else:
        err_buf.error("Bad usage: %s, command: '%s %s'" % (msg, cmd, args))


def empty_cib_err():
    err_buf.error("No CIB!")


def cib_parse_err(msg, s):
    err_buf.error("%s" % msg)
    err_buf.info("offending string: %s" % s)


def cib_ver_unsupported_err(validator, rel):
    err_buf.error("Unsupported CIB: validator '%s', release '%s'" %
                  (validator, rel))
    err_buf.error("To upgrade an old (<1.0) schema, use the upgrade command.")


def update_err(obj_id, cibadm_opt, xml, rc):
    if cibadm_opt == '-U':
        task = "update"
    elif cibadm_opt == '-D':
        task = "delete"
    elif cibadm_opt == '-P':
        task = "patch"
    else:
        task = "replace"
    err_buf.error("could not %s %s (rc=%d)" % (task, obj_id, rc))
    if rc == 54:
        err_buf.info("Permission denied.")
    elif task == "patch":
        err_buf.info("offending xml diff: %s" % xml)
    else:
        err_buf.info("offending xml: %s" % xml)


err_buf = ErrorBuffer()
# vim:ts=4:sw=4:et:



# Copyright (C) 2016 Kristoffer Gronlund <kgronlund@suse.com>
# See COPYING for license information.

import bz2
import gzip
import re
import os
import sys
import collections
import json
import time

from . import msg as crmlog
from . import xmlutil
from . import logtime
from . import utils
from . import log_patterns


_METADATA_FILENAME = "__meta.json"
_METADATA_CACHE_AGE = (60.0 * 60.0)
# Update this when changing the metadata format
_METADATA_VERSION = 1


def _open_logfile(logfile):
    """
    Open a file which may be gz|bz2 compressed.
    Uncompress based on extension.
    """
    try:
        if logfile.endswith(".bz2"):
            return bz2.BZ2File(logfile)
        if logfile.endswith(".gz"):
            return gzip.open(logfile)
        return open(logfile, "rb")
    except IOError as msg:
        crmlog.common_error("open %s: %s" % (logfile, msg))
        return None


def _transition_start_re():
    """
    Return regular expression matching transition start.
    number_re can be a specific transition or a regexp matching
    any transition number.
    The resulting RE has groups
    1: transition number
    2: full path of pe file
    3: pe file number
    """
    m1 = "crmd.*Processing graph ([0-9]+).*derived from (.*/pe-[^-]+-([0-9]+)[.]bz2)"
    m2 = "pengine.*[Tt]ransition ([0-9]+).*([^ ]*/pe-[^-]+-([0-9]+)[.]bz2)"
    try:
        return re.compile("(?:%s)|(?:%s)" % (m1, m2))
    except re.error as e:
        crmlog.common_debug("RE compilation failed: %s" % (e))
        raise ValueError("Error in search expression")


def pefile_shortname(pe_file):
    return os.path.basename(pe_file).replace(".bz2", "")


def trans_str(node, pe_file):
    '''Convert node,pe_file to transition string.'''
    return "%s:%s" % (node, pefile_shortname(pe_file))


def _transition_end_re():
    """
    Return RE matching transition end.
    See transition_start_re for more details.

    1: trans_num
    2: pe_file
    3: pe_num
    4: state
    """
    try:
        return re.compile("crmd.*Transition ([0-9]+).*Source=(.*/pe-[^-]+-([0-9]+)[.]bz2).:.*(Stopped|Complete|Terminated)")
    except re.error as e:
        crmlog.common_debug("RE compilation failed: %s" % (e))
        raise ValueError("Error in search expression")


_GRAPH_ACTIONS_RE = re.compile("([A-Z][a-z]+)=([0-9]+)")


def _run_graph_msg_actions(msg):
    '''
    crmd: [13667]: info: run_graph: Transition 399 (Complete=5,
    Pending=1, Fired=1, Skipped=0, Incomplete=3,
    Source=...
    Returns dict: d[Pending]=np, d[Fired]=nf, ...
    Only stores non-zero values.
    '''
    d = {}
    s = msg
    r = _GRAPH_ACTIONS_RE.search(s)
    while r:
        val = int(r.group(2))
        if val != 0:
            d[r.group(1)] = val
        s = s[r.end():]
        r = _GRAPH_ACTIONS_RE.search(s)
    return d


def mk_re_list(patt_l, repl):
    'Build a list of regular expressions, replace "%%" with repl'
    l = []
    for re_l in patt_l:
        l += [x.replace("%%", repl) for x in re_l]
    if not repl:
        l = [x.replace(".*.*", ".*") for x in l]
    return l


class Transition(object):
    __slots__ = ('loc', 'dc', 'start_ts', 'trans_num', 'pe_file', 'pe_num', 'end_ts', 'end_state', 'end_actions', 'tags')

    def __init__(self, loc, dc, start_ts, trans_num, pe_file, pe_num):
        self.loc = loc
        self.dc = dc
        self.start_ts = start_ts
        self.trans_num = trans_num
        self.pe_file = pe_file
        self.pe_num = pe_num
        self.end_ts = None
        self.end_state = None
        self.end_actions = None
        self.tags = set()

    def __str__(self):
        return trans_str(self.dc, self.pe_file)

    def shortname(self):
        return pefile_shortname(self.pe_file)

    def actions(self):
        return self.end_actions

    def actions_count(self):
        if self.end_actions is not None:
            return sum(self.end_actions.values())
        return -1

    def path(self):
        return os.path.join(self.loc, self.dc, "pengine", self.pe_file)

    def description(self):
        s = "%s %s - %s: %-15s %-15s %s" % (
            utils.shortdate(self.start_ts),
            utils.shorttime(self.start_ts),
            self.end_ts and utils.shorttime(self.end_ts) or "--:--:--",
            self.shortname(),
            self.dc,
            " ".join(sorted(self.tags))
        )
        return s

    def empty(self, prev):
        """
        True if this transition resulted in no actions and no CIB changes
        prev: previous transition
        """
        old_pe_l_file = prev.path()
        new_pe_l_file = self.path()
        no_actions = self.actions_count() == 0
        if not os.path.isfile(old_pe_l_file) or not os.path.isfile(new_pe_l_file):
            return no_actions
        old_cib = xmlutil.compressed_file_to_cib(old_pe_l_file)
        new_cib = xmlutil.compressed_file_to_cib(new_pe_l_file)
        if old_cib is None or new_cib is None:
            return no_actions
        prev_epoch = old_cib.attrib.get("epoch", "0")
        epoch = new_cib.attrib.get("epoch", "0")
        prev_admin_epoch = old_cib.attrib.get("admin_epoch", "0")
        admin_epoch = new_cib.attrib.get("admin_epoch", "0")
        return no_actions and epoch == prev_epoch and admin_epoch == prev_admin_epoch

    def transition_info(self):
        print("Transition %s (%s -" % (self, utils.shorttime(self.start_ts)), end=' ')
        if self.end_ts:
            print("%s):" % utils.shorttime(self.end_ts))
            act_d = self.actions()
            total = self.actions_count()
            s = ", ".join(["%d %s" % (act_d[x], x) for x in act_d if act_d[x]])
            print("\ttotal %d actions: %s" % (total, s))
        else:
            print("[unfinished])")

    def to_dict(self):
        """
        Serialize to dict (for cache)
        """
        o = {"tags": list(self.tags)}
        for k in self.__slots__:
            if k in ("loc", "tags"):
                continue
            o[k] = getattr(self, k)
        return o

    @classmethod
    def from_dict(cls, loc, obj):
        t = Transition(loc, None, None, None, None, None)
        for k, v in obj.items():
            setattr(t, k, set(v) if k == "tags" else v)
        return t


class CibInfo(object):
    def __init__(self, report_path):
        self.filename = utils.file_find_by_name(report_path, "cib.xml")
        self.nodes = []
        self.primitives = []
        self.groups = {}
        self.clones = {}
        self.cloned_resources = set()
        self.not_cloned_resources = set()

        cib_elem = None
        if self.filename:
            cib_elem = xmlutil.file2cib_elem(self.filename)

        if cib_elem is None:
            self.nodes = utils.list_cluster_nodes()
            return

        self.nodes = [x.get("uname") or x.get("id") for x in cib_elem.xpath("/cib/configuration/nodes/node")]

        self.primitives = [x.get("id") for x in cib_elem.xpath("/cib/configuration/resources//primitive")]

        for grp in cib_elem.xpath("/cib/configuration/resources/group"):
            self.groups[grp.get("id")] = xmlutil.get_rsc_children_ids(grp)

        for cln in cib_elem.xpath("/cib/configuration/resources/*[self::clone or self::master]"):
            self.clones[cln.get("id")] = xmlutil.get_prim_children_ids(cln)
            self.cloned_resources.union(self.clones[cln.get("id")])

        self.not_cloned_resources = set(x for x in self.primitives if x not in self.cloned_resources)

    def resources(self):
        return self.primitives + list(self.groups.keys()) + list(self.clones.keys())

    def match_resources(self):
        """
        list of regex expressions to match resources
        """
        rsc_l = list(self.not_cloned_resources)
        rsc_l += ["%s(?::[0-9]+)?" % x for x in self.cloned_resources]
        return rsc_l


class LogParser(object):
    """
    Used by the history explorer.
    Given a report directory, generates log metadata.

    TODO:

    This information is then written to a file called %(_METADATA_FILENAME),
    and the next time the history explorer is started, we skip the
    analysis and load the metadata directly.

    The analysis is done over the complete log: Timeframe narrowing happens elsewhere.
    """

    def __init__(self, loc, cib, logfiles, detail):
        """
        report_root: Base directory of the report
        """
        self.loc = loc
        self.cib = cib
        self.filenames = logfiles
        self.fileobjs = [_open_logfile(f) for f in logfiles]
        self.detail = detail

        self.events = {}
        self.transitions = []

        self.from_ts = None
        self.to_ts = None

    def __del__(self):
        for f in self.fileobjs:
            f.close()

    def scan(self, mode=None):
        """
        mode = 'refresh':
        Re-read logs that may have new data appended.
        Right now this re-scans all the log data.
        TODO: Only scan new data by tracking the previous
        end of each file and scanning from there. Retain
        previous data and just add new transitions / events.

        Returns list of pefiles missing from report. [(node, [pefile ...]) ...]

        mode = 'force':
        Completely re-parse (ignore cache)
        """
        with utils.nogc():
            return self._scan(mode=mode)

    def _scan(self, mode):
        """
        Scan logs and generate metadata for transitions,
        tags and events. (used when retreiving log lines later)

        Returns list of pefiles missing from report. [(node, [pefile ...]) ...]

        mode: None, 'refresh' or 'force'

        TODO: Load/save metadata when already generated.
        TODO: scan each logfile in a separate thread?
        """

        if mode not in ('refresh', 'force') and self._load_cache():
            return []

        missing_pefiles = []

        # {etype -> [(sortkey, msg)]}
        # TODO: store (sortkey, fileid, spos) instead?
        self.events = collections.defaultdict(list)

        self.transitions = []

        # trans_num:pe_num -> Transition()
        transitions_map = {}

        startre = _transition_start_re()
        endre = _transition_end_re()

        eventre = {}
        eventre["node"] = self._build_re("node", self.cib.nodes)
        eventre["resource"] = self._build_re("resource", self.cib.match_resources())
        eventre["quorum"] = self._build_re("quorum", [])
        eventre["events"] = self._build_re("events", [])

        DEFAULT, IN_TRANSITION = 0, 1
        state = DEFAULT
        transition = None

        for logidx, (filename, log) in enumerate(zip(self.filenames, self.fileobjs)):
            log.seek(0)
            crmlog.common_debug("parsing %s" % (filename))
            line = "a"
            while line != '':
                spos = log.tell()
                line = utils.to_ascii(log.readline())
                m = startre.search(line)
                if m:
                    # m.groups() is (transnum1, pefile1, penum1, transnum2, pefile2, penum2) where
                    # it matched either 1 or 2
                    t1, p1, n1, t2, p2, n2 = m.groups()
                    if t1 is not None:
                        trans_num, pe_file, pe_num = t1, p1, n1
                    else:
                        trans_num, pe_file, pe_num = t2, p2, n2
                    pe_orig = pe_file
                    pe_file = os.path.basename(pe_orig)
                    ts, dc = logtime.syslog_ts_node(line)
                    if ts is None or dc is None:
                        continue
                    id_ = trans_str(dc, pe_file)
                    transition = transitions_map.get(id_)
                    if transition is None:
                        transition = Transition(self.loc, dc, ts, trans_num, pe_file, pe_num)
                        self.transitions.append(transition)
                        transitions_map[id_] = transition
                        crmlog.common_debug("{Transition: %s" % (transition))

                        if not os.path.isfile(transition.path()):
                            missing_pefiles.append((dc, pe_orig))
                    else:
                        crmlog.common_debug("~Transition: %s old(%s, %s) new(%s, %s)" %
                                            (transition, transition.trans_num, transition.pe_file, trans_num, pe_file))
                    state = IN_TRANSITION
                    continue
                if state == IN_TRANSITION:
                    m = endre.search(line)
                    if m:
                        trans_num, pe_file, pe_num, state = m.groups()
                        pe_file = os.path.basename(pe_file)
                        ts, dc = logtime.syslog_ts_node(line)
                        if ts is None or dc is None:
                            continue
                        transition = transitions_map.get(trans_str(dc, pe_file))
                        if transition is None:
                            # transition end without previous begin...
                            crmlog.common_debug("Found transition end without start: %s: %s - %s:%s" % (ts, filename, trans_num, pe_file))
                        else:
                            transition.end_state = state
                            transition.end_ts = ts
                            transition.end_actions = _run_graph_msg_actions(line)
                            crmlog.common_debug("}Transition: %s %s" % (transition, state))
                        state = DEFAULT

                # events
                for etype, erx in eventre.items():
                    for rx in erx:
                        m = rx.search(line)
                        if m:
                            ts = logtime.syslog_ts(line)
                            if ts is None:
                                continue
                            crmlog.common_debug("+Event %s: %s" % (etype, ", ".join(m.groups())))
                            sk = (int(ts) << 32) + int(spos)
                            self.events[etype].append((sk, logidx, spos))
                            if transition is not None:
                                for t in m.groups():
                                    if t:
                                        transition.tags.add(t.lower())

                if state == DEFAULT:
                    transition = None

        self.transitions.sort(key=lambda t: t.start_ts)
        for etype, logs in self.events.items():
            logs.sort(key=lambda e: e[0])
        empties = []
        for i, t in enumerate(self.transitions):
            if i == 0:
                continue
            if t.empty(self.transitions[i - 1]):
                empties.append(t)
        self.transitions = [t for t in self.transitions if t not in empties]
        self._save_cache()
        if missing_pefiles:
            rdict = collections.defaultdict(list)
            for node, pe in missing_pefiles:
                rdict[node].append(pe)
            missing_pefiles = list(rdict.items())
        return missing_pefiles

    def set_timeframe(self, from_t, to_t):
        """
        from_t, to_t: timestamps or datetime objects
        """
        self.from_ts = logtime.make_time(from_t)
        self.to_ts = logtime.make_time(to_t)

    def get_logs(self, nodes=None):
        """
        Generator which yields a list of log messages limited by the
        list of nodes, or from all nodes.

        The log lines are printed in order, by reading from
        all files at once and always printing the line with
        the lowest timestamp
        """

        def include_log(logfile):
            return not nodes or os.path.basename(os.path.dirname(logfile)) in nodes

        for f in self.fileobjs:
            f.seek(0)

        lines = [[None, utils.to_ascii(f.readline()), f] for f in self.fileobjs]
        for i, line in enumerate(lines):
            if not line[1]:
                line[0], line[2] = sys.float_info.max, None
            else:
                line[0] = logtime.syslog_ts(line[1])

        while any(f is not None for _, _, f in lines):
            x = min(lines, key=lambda v: v[0])
            if x[0] is None or x[2] is None:
                break
            if self.to_ts and x[0] > self.to_ts:
                break
            if not (self.from_ts and x[0] < self.from_ts):
                yield x[1]
            x[1] = utils.to_ascii(x[2].readline())
            if not x[1]:
                x[0], x[2] = sys.float_info.max, None
            else:
                x[0] = logtime.syslog_ts(x[1])

    def get_events(self, event=None, nodes=None, resources=None):
        """
        Generator which outputs matching event lines
        event: optional node, resource, quorum
        nodes: optional list of nodes
        resources: optional list of resources

        TODO: ordering, time limits
        """
        if event is not None:
            eventlogs = [event]
        else:
            eventlogs = sorted(list(self.events.keys()))

        if nodes:
            rxes = self._build_re(event, nodes)
        elif resources:
            expanded_l = []
            for r in resources:
                if r in self.cib.groups:
                    expanded_l += self.cib.groups[r]
                elif r in self.cib.clones:
                    expanded_l += self.cib.clones[r]
                else:
                    expanded_l.append(r)

            def clonify(r):
                return r + "(?::[0-9]+)?" if r in self.cib.cloned_resources else r
            expanded_l = [clonify(r) for r in expanded_l]
            rxes = self._build_re(event, expanded_l)
        else:
            rxes = None

        if event == "resource" and resources is not None and rxes is not None:
            crmlog.common_debug("resource %s rxes: %s" % (", ".join(resources), ", ".join(r.pattern for r in rxes)))

        if rxes is not None:
            for log in eventlogs:
                for _, f, pos in self.events.get(log, []):
                    self.fileobjs[f].seek(pos)
                    msg = utils.to_ascii(self.fileobjs[f].readline())
                    if any(rx.search(msg) for rx in rxes):
                        ts = logtime.syslog_ts(msg)
                        if not (self.from_ts and ts < self.from_ts) and not (self.to_ts and ts > self.to_ts):
                            yield msg
        else:
            for log in eventlogs:
                for _, f, pos in self.events.get(log, []):
                    self.fileobjs[f].seek(pos)
                    msg = utils.to_ascii(self.fileobjs[f].readline())
                    ts = logtime.syslog_ts(msg)
                    if not (self.from_ts and ts < self.from_ts) and not (self.to_ts and ts > self.to_ts):
                        yield msg

    def get_transitions(self):
        """
        Yields transitions within the current timeframe
        """
        for t in self.transitions:
            if not (self.from_ts and t.end_ts and t.end_ts < self.from_ts) and not (self.to_ts and t.start_ts and t.start_ts > self.to_ts):
                yield t

    def _get_patt_l(self, etype):
        '''
        get the list of patterns for this type, up to and
        including current detail level
        '''
        patterns = log_patterns.patterns(cib_f=self.cib.filename)
        if etype not in patterns:
            crmlog.common_error("%s not featured in log patterns" % etype)
            return None
        return patterns[etype][0:self.detail+1]

    def _build_re(self, etype, args):
        '''
        Prepare a regex string for the type and args.
        For instance, "resource" and rsc1, rsc2, ...
        '''
        patt_l = self._get_patt_l(etype)
        if not patt_l:
            return None
        if not args:
            re_l = mk_re_list(patt_l, "")
        else:
            re_l = mk_re_list(patt_l, r'(%s)' % "|".join(args))
        return [re.compile(r) for r in re_l]

    def to_dict(self):
        """
        Serialize self to dict (including transition objects)
        """
        o = {
            "version": _METADATA_VERSION,
            "events": self.events,
            "transitions": [t.to_dict() for t in self.transitions],
            "cib": {
                "nodes": self.cib.nodes,
                "primitives": self.cib.primitives,
                "groups": self.cib.groups,
                "clones": self.cib.clones
            }
        }
        return o

    def from_dict(self, obj):
        """
        Load from dict
        """
        if "version" not in obj or obj["version"] != _METADATA_VERSION:
            return False
        self.events = obj["events"]
        self.transitions = [Transition.from_dict(self.loc, t) for t in obj["transitions"]]
        return True

    def _metafile(self):
        return os.path.join(self.loc, _METADATA_FILENAME)

    def count(self):
        """
        Returns (num transitions, num events)
        """
        return len(self.transitions), sum(len(e) for e in list(self.events.values()))

    def _save_cache(self):
        """
        Save state to cache file
        """
        fn = self._metafile()
        try:
            with open(fn, 'wt') as f:
                json.dump(self.to_dict(), f, indent=2)
                crmlog.common_debug("Transition metadata saved to %s" % (fn))
        except IOError as e:
            crmlog.common_debug("Could not update metadata cache: %s" % (e))

    def _load_cache(self):
        """
        Load state from cache file
        """
        fn = self._metafile()
        if os.path.isfile(fn):
            meta_mtime = os.stat(fn).st_mtime
            logf_mtime = max([os.stat(f).st_mtime for f in self.filenames if os.path.isfile(f)])

            if meta_mtime >= logf_mtime and time.time() - meta_mtime < _METADATA_CACHE_AGE:
                try:
                    with open(fn, 'r') as f:
                        try:
                            if not self.from_dict(json.load(f)):
                                return False
                            crmlog.common_debug("Transition metadata loaded from %s" % (fn))
                            return True
                        except ValueError as e:
                            crmlog.common_debug("Failed to load metadata: %s" % (e))
                except IOError as e:
                    return False
        return False
